"""Tests for Sentence Transformers embedding adapter."""

from unittest.mock import AsyncMock, MagicMock, patch

import numpy as np
import pytest

from acb.adapters.embedding._base import EmbeddingBatch
from acb.adapters.embedding.sentence_transformers import (
    SentenceTransformersEmbedding,
    SentenceTransformersSettings,
)
from acb.depends import depends


@pytest.fixture
def mock_sentence_transformer():
    """Mock SentenceTransformer model."""
    model = MagicMock()
    model.max_seq_length = 512
    model.get_sentence_embedding_dimension.return_value = 384

    # Mock encode method
    def mock_encode(texts, **kwargs):
        embeddings = []
        for i, text in enumerate(texts):
            # Create deterministic embeddings
            embedding = np.array([0.1 + i * 0.1] * 384, dtype=np.float32)
            if kwargs.get("normalize_embeddings", False):
                embedding = embedding / np.linalg.norm(embedding)
            embeddings.append(embedding)
        return np.array(embeddings)

    model.encode = mock_encode
    return model


@pytest.fixture
def mock_config():
    """Mock configuration."""
    config = MagicMock()
    # Add necessary attributes that the sentence transformers adapter expects
    config.debug = MagicMock()
    config.app = MagicMock()
    config.app.name = "test_app"
    config.embedding = MagicMock()
    config.embedding.model = "all-MiniLM-L6-v2"
    config.embedding.device = "cpu"
    return config


@pytest.fixture
async def sentence_transformers_adapter(mock_config, mock_sentence_transformer):
    """Create Sentence Transformers embedding adapter with mocked dependencies."""
    with (
        patch(
            "acb.adapters.embedding.sentence_transformers._sentence_transformers_available",
            True,
        ),
        patch(
            "acb.adapters.embedding.sentence_transformers.SentenceTransformer"
        ) as mock_st_class,
        patch("acb.adapters.embedding.sentence_transformers.torch") as mock_torch,
        patch.object(depends, "get") as mock_depends,
    ):
        mock_depends.return_value = mock_config
        mock_torch.cuda.is_available.return_value = False
        mock_st_class.return_value = mock_sentence_transformer

        settings = SentenceTransformersSettings(
            model="all-MiniLM-L6-v2",
            device="cpu",
        )

        adapter = SentenceTransformersEmbedding(settings)
        adapter._model = mock_sentence_transformer
        adapter._device = "cpu"

        yield adapter, mock_sentence_transformer

        await adapter.cleanup()


@pytest.mark.asyncio
class TestSentenceTransformersEmbedding:
    """Test cases for Sentence Transformers embedding adapter."""

    async def test_initialization(self, mock_config):
        """Test adapter initialization."""
        with (
            patch(
                "acb.adapters.embedding.sentence_transformers._sentence_transformers_available",
                True,
            ),
            patch.object(depends, "get", return_value=mock_config),
        ):
            settings = SentenceTransformersSettings(
                model="all-MiniLM-L6-v2",
                device="cpu",
            )

            adapter = SentenceTransformersEmbedding(settings)
            assert adapter._settings.model == "all-MiniLM-L6-v2"
            assert adapter._settings.device == "cpu"

    async def test_initialization_without_sentence_transformers(self):
        """Test adapter initialization when Sentence Transformers is not available."""
        with patch(
            "acb.adapters.embedding.sentence_transformers._sentence_transformers_available",
            False,
        ):
            with pytest.raises(
                ImportError, match="Sentence Transformers library not available"
            ):
                SentenceTransformersEmbedding()

    async def test_load_model_cpu(self, mock_config):
        """Test model loading on CPU."""
        with (
            patch(
                "acb.adapters.embedding.sentence_transformers._sentence_transformers_available",
                True,
            ),
            patch("acb.adapters.embedding.sentence_transformers.SentenceTransformer"),
            patch("acb.adapters.embedding.sentence_transformers.torch") as mock_torch,
            patch.object(depends, "get", return_value=mock_config),
            patch("asyncio.get_event_loop") as mock_loop,
        ):
            mock_torch.cuda.is_available.return_value = False
            mock_executor = AsyncMock()
            mock_loop.return_value.run_in_executor = mock_executor

            settings = SentenceTransformersSettings(device="auto")
            adapter = SentenceTransformersEmbedding(settings)

            await adapter._load_model()

            assert adapter._device == "cpu"

    async def test_load_model_cuda(self, mock_config):
        """Test model loading on CUDA."""
        with (
            patch(
                "acb.adapters.embedding.sentence_transformers._sentence_transformers_available",
                True,
            ),
            patch("acb.adapters.embedding.sentence_transformers.SentenceTransformer"),
            patch("acb.adapters.embedding.sentence_transformers.torch") as mock_torch,
            patch.object(depends, "get", return_value=mock_config),
            patch("asyncio.get_event_loop") as mock_loop,
        ):
            mock_torch.cuda.is_available.return_value = True
            mock_executor = AsyncMock()
            mock_loop.return_value.run_in_executor = mock_executor

            settings = SentenceTransformersSettings(device="auto")
            adapter = SentenceTransformersEmbedding(settings)

            await adapter._load_model()

            assert adapter._device == "cuda"

    async def test_embed_texts_success(self, sentence_transformers_adapter):
        """Test successful text embedding."""
        adapter, mock_model = sentence_transformers_adapter

        texts = ["Hello world", "How are you?"]
        batch = await adapter._embed_texts(
            texts=texts,
            model="all-MiniLM-L6-v2",
            normalize=True,
            batch_size=10,
        )

        assert isinstance(batch, EmbeddingBatch)
        assert len(batch.results) == 2
        assert batch.model == "all-MiniLM-L6-v2"

        # Check individual results
        for i, result in enumerate(batch.results):
            assert result.text == texts[i]
            assert len(result.embedding) == 384
            assert result.model == "all-MiniLM-L6-v2"
            assert result.dimensions == 384
            assert result.metadata["device"] == "cpu"
            assert result.metadata["normalized"] is True

    async def test_embed_texts_without_normalization(
        self, sentence_transformers_adapter
    ):
        """Test text embedding without normalization."""
        adapter, mock_model = sentence_transformers_adapter

        texts = ["Test text"]
        batch = await adapter._embed_texts(
            texts=texts,
            model="all-MiniLM-L6-v2",
            normalize=False,
            batch_size=1,
        )

        result = batch.results[0]
        assert result.metadata["normalized"] is False

    async def test_embed_documents(self, sentence_transformers_adapter):
        """Test document embedding with chunking."""
        adapter, mock_model = sentence_transformers_adapter

        documents = ["This is a long document that should be chunked. " * 20]
        batches = await adapter._embed_documents(
            documents=documents,
            chunk_size=100,
            chunk_overlap=20,
            model="all-MiniLM-L6-v2",
        )

        assert len(batches) == 1
        batch = batches[0]
        assert len(batch.results) > 1  # Should be chunked

        # Check metadata
        for result in batch.results:
            assert result.metadata["is_chunk"] is True
            assert "document_id" in result.metadata

    async def test_similarity_search(self, sentence_transformers_adapter):
        """Test semantic similarity search."""
        adapter, mock_model = sentence_transformers_adapter

        # Mock similarity method
        def mock_similarity(query_embeddings, doc_embeddings):
            # Return mock similarities
            return np.array([[0.9, 0.7, 0.5]])

        mock_model.similarity = mock_similarity

        query = "test query"
        documents = ["similar document", "somewhat similar", "different text"]

        results = await adapter.similarity_search(query, documents, top_k=2)

        assert len(results) == 2
        assert results[0][1] > results[1][1]  # Should be sorted by similarity

    async def test_compute_similarity(self, sentence_transformers_adapter):
        """Test similarity computation."""
        adapter, _ = sentence_transformers_adapter

        embedding1 = [1.0, 0.0, 0.0]
        embedding2 = [0.0, 1.0, 0.0]

        # Test cosine similarity
        similarity = await adapter._compute_similarity(embedding1, embedding2, "cosine")
        assert similarity == 0.0  # Orthogonal vectors

        # Test unsupported method
        with pytest.raises(ValueError, match="Unsupported similarity method"):
            await adapter._compute_similarity(embedding1, embedding2, "unsupported")

    async def test_get_model_info(self, sentence_transformers_adapter):
        """Test getting model information."""
        adapter, mock_model = sentence_transformers_adapter

        info = await adapter._get_model_info("all-MiniLM-L6-v2")

        assert info["name"] == "all-MiniLM-L6-v2"
        assert info["provider"] == "sentence_transformers"
        assert info["device"] == "cpu"
        assert info["local"] is True
        assert info["max_position_embeddings"] is None  # Mock doesn't have this
        assert info["hidden_size"] is None  # Mock doesn't have this

    async def test_list_models(self, sentence_transformers_adapter):
        """Test listing available models."""
        adapter, _ = sentence_transformers_adapter

        models = await adapter._list_models()
        assert isinstance(models, list)
        assert len(models) > 0

        # Check model structure
        for model in models:
            assert "name" in model
            assert "provider" in model
            assert "description" in model
            assert model["provider"] == "sentence_transformers"

    async def test_health_check_success(self, sentence_transformers_adapter):
        """Test successful health check."""
        adapter, mock_model = sentence_transformers_adapter

        health = await adapter.health_check()

        assert health["status"] == "healthy"
        assert health["provider"] == "sentence_transformers"
        assert health["model"] == "all-MiniLM-L6-v2"
        assert health["device"] == "cpu"
        assert "timestamp" in health

    async def test_health_check_failure(self, sentence_transformers_adapter):
        """Test health check failure."""
        adapter, mock_model = sentence_transformers_adapter

        # Mock embed_text to raise an exception
        with patch.object(adapter, "embed_text", side_effect=Exception("Model error")):
            health = await adapter.health_check()

            assert health["status"] == "unhealthy"
            assert health["provider"] == "sentence_transformers"
            assert "error" in health
            assert "timestamp" in health

    async def test_cleanup(self, sentence_transformers_adapter):
        """Test adapter cleanup."""
        adapter, mock_model = sentence_transformers_adapter

        with patch("acb.adapters.embedding.sentence_transformers.torch") as mock_torch:
            mock_torch.cuda.is_available.return_value = True
            mock_torch.cuda.empty_cache = MagicMock()

            await adapter.cleanup()

            # Model should be cleaned up
            assert adapter._model is None
            assert adapter._tokenizer is None

            # CUDA cache should be cleared
            mock_torch.cuda.empty_cache.assert_called_once()

    async def test_error_handling(self, sentence_transformers_adapter):
        """Test error handling during embedding generation."""
        adapter, mock_model = sentence_transformers_adapter

        # Mock encode to raise an exception
        mock_model.encode = MagicMock(side_effect=Exception("Encoding error"))

        with pytest.raises(Exception, match="Encoding error"):
            await adapter._embed_texts(
                texts=["test"],
                model="all-MiniLM-L6-v2",
                normalize=True,
                batch_size=1,
            )

    async def test_precision_settings(self, mock_config):
        """Test model precision settings."""
        with (
            patch(
                "acb.adapters.embedding.sentence_transformers._sentence_transformers_available",
                True,
            ),
            patch(
                "acb.adapters.embedding.sentence_transformers.SentenceTransformer"
            ) as mock_st_class,
            patch.object(depends, "get", return_value=mock_config),
        ):
            mock_model = MagicMock()
            mock_model.half = MagicMock()
            mock_st_class.return_value = mock_model

            settings = SentenceTransformersSettings(
                device="cuda",
                precision="float16",
            )

            adapter = SentenceTransformersEmbedding(settings)
            adapter._device = "cuda"

            await adapter._load_model()

            # half() should be called for float16 precision on CUDA
            mock_model.half.assert_called_once()


@pytest.mark.asyncio
class TestSentenceTransformersSettings:
    """Test cases for Sentence Transformers settings."""

    def test_settings_defaults(self):
        """Test default settings values."""
        settings = SentenceTransformersSettings()

        assert settings.model == "all-MiniLM-L6-v2"
        assert settings.device == "auto"
        assert settings.normalize_embeddings is True
        assert settings.convert_to_numpy is True
        assert settings.convert_to_tensor is False
        assert settings.batch_size == 32

    def test_settings_custom_values(self):
        """Test custom settings values."""
        settings = SentenceTransformersSettings(
            model="all-mpnet-base-v2",
            device="cuda",
            batch_size=64,
            normalize_embeddings=False,
        )

        assert settings.model == "all-mpnet-base-v2"
        assert settings.device == "cuda"
        assert settings.batch_size == 64
        assert settings.normalize_embeddings is False

    def test_environment_prefix(self):
        """Test environment variable prefix."""
        assert (
            SentenceTransformersSettings.Config.env_prefix == "SENTENCE_TRANSFORMERS_"
        )


@pytest.mark.asyncio
class TestSentenceTransformersIntegration:
    """Integration tests for Sentence Transformers adapter."""

    async def test_end_to_end_workflow(self, sentence_transformers_adapter):
        """Test complete embedding workflow."""
        adapter, mock_model = sentence_transformers_adapter

        # Single text embedding
        text = "Test embedding workflow"
        embedding = await adapter.embed_text(text)
        assert len(embedding) == 384

        # Multiple texts
        texts = ["First text", "Second text", "Third text"]
        batch = await adapter.embed_texts(texts)
        assert len(batch.results) == 3

        # Document chunking
        long_document = "This is a very long document. " * 50
        doc_batches = await adapter.embed_documents([long_document])
        assert len(doc_batches) == 1
        assert len(doc_batches[0].results) > 1

        # Model information
        info = await adapter.get_model_info()
        assert "name" in info

        # List models
        models = await adapter.list_models()
        assert len(models) > 0



"""OpenAI embeddings adapter implementation."""

import time

import asyncio
import typing as t
from datetime import datetime
from pydantic import Field

from acb.adapters import (
    AdapterCapability,
    AdapterMetadata,
    AdapterStatus,
    generate_adapter_id,
)
from acb.adapters.embedding._base import (
    EmbeddingAdapter,
    EmbeddingBaseSettings,
    EmbeddingBatch,
    EmbeddingModel,
    EmbeddingResult,
    EmbeddingUtils,
)
from acb.config import Config
from acb.depends import depends

if t.TYPE_CHECKING:
    from openai import AsyncOpenAI
else:
    try:
        from openai import AsyncOpenAI  # type: ignore[no-redef]

        _openai_available = True
    except ImportError:
        AsyncOpenAI = None  # type: ignore[assignment,misc,no-redef]
        _openai_available = False

MODULE_METADATA = AdapterMetadata(
    module_id=generate_adapter_id(),
    name="OpenAI Embeddings",
    category="embedding",
    provider="openai",
    version="1.0.0",
    acb_min_version="0.19.0",
    author="ACB Development Team",
    created_date="2025-09-30",
    last_modified="2025-09-30",
    status=AdapterStatus.STABLE,
    capabilities=[
        AdapterCapability.ASYNC_OPERATIONS,
        AdapterCapability.BATCHING,
        AdapterCapability.CACHING,
        AdapterCapability.TLS_SUPPORT,
        AdapterCapability.METRICS,
        AdapterCapability.BATCH_EMBEDDING,
        AdapterCapability.VECTOR_NORMALIZATION,
        AdapterCapability.TEXT_PREPROCESSING,
    ],
    required_packages=["openai>=1.0.0"],
    description="High-performance OpenAI embeddings adapter with batch processing and caching",
    settings_class="OpenAIEmbeddingSettings",
)


class OpenAIEmbeddingSettings(EmbeddingBaseSettings):
    """OpenAI-specific embedding settings."""

    # OpenAI-specific settings
    organization: str | None = Field(default=None, description="OpenAI organization ID")
    dimensions: int | None = Field(
        default=None,
        description="Embedding dimensions (for v3 models)",
    )
    encoding_format: str = Field(
        default="float",
        description="Encoding format (float or base64)",
    )

    # Rate limiting
    requests_per_minute: int = Field(default=3000)
    tokens_per_minute: int = Field(default=1000000)

    # Override base settings defaults
    model: str = Field(default=EmbeddingModel.TEXT_EMBEDDING_3_SMALL.value)
    batch_size: int = Field(default=100)  # OpenAI supports up to 2048 inputs

    class Config:
        env_prefix = "OPENAI_"


class OpenAIEmbedding(EmbeddingAdapter):
    """OpenAI embeddings adapter implementation."""

    def __init__(self, settings: OpenAIEmbeddingSettings | None = None) -> None:
        if not _openai_available:
            msg = "OpenAI library not available. Install with: pip install openai"
            raise ImportError(
                msg,
            )

        if settings is None:
            settings = OpenAIEmbeddingSettings()

        super().__init__(settings)
        self._settings = settings
        self._client: t.Any = None
        self._rate_limiter = None
        self._last_request_time = 0.0

    async def _ensure_client(self) -> t.Any:
        """Ensure OpenAI client is initialized."""
        if self._client is None:
            if AsyncOpenAI is None:
                msg = "OpenAI library not available"
                raise ImportError(msg)
            # Cast settings to access OpenAI-specific fields
            settings = t.cast(OpenAIEmbeddingSettings, self._settings)
            self._client = AsyncOpenAI(  # type: ignore[assignment]
                api_key=self._settings.api_key.get_secret_value()
                if self._settings.api_key
                else "",
                organization=settings.organization,
                base_url=self._settings.base_url,
                timeout=self._settings.timeout,
                max_retries=self._settings.max_retries,
            )
            self.register_resource(self._client)

        return self._client

    async def _embed_texts(
        self,
        texts: list[str],
        model: str,
        normalize: bool,
        batch_size: int,
        **kwargs: t.Any,
    ) -> EmbeddingBatch:
        """Generate embeddings for multiple texts using OpenAI."""
        start_time = time.time()
        client = await self._ensure_client()
        logger: t.Any = depends.get("logger")

        results = []
        total_tokens = 0

        # Process texts in batches
        batches = self._batch_texts(texts, batch_size)

        for batch_texts in batches:
            try:
                await self._apply_rate_limit()

                # Process single batch
                batch_results, batch_tokens = await self._process_embedding_batch(
                    client,
                    batch_texts,
                    model,
                    normalize,
                )

                results.extend(batch_results)
                total_tokens += batch_tokens

                await logger.debug(
                    f"OpenAI embeddings batch completed: {len(batch_texts)} texts, model: {model}",
                )

            except Exception as e:
                await logger.exception(f"Error generating OpenAI embeddings: {e}")
                raise

        processing_time = time.time() - start_time

        return EmbeddingBatch(
            results=results,
            total_tokens=total_tokens if total_tokens > 0 else None,
            processing_time=processing_time,
            model=model,
            batch_size=len(results),
        )

    async def _process_embedding_batch(
        self,
        client: t.Any,
        batch_texts: list[str],
        model: str,
        normalize: bool,
    ) -> tuple[list[EmbeddingResult], int]:
        """Process a single batch of embeddings.

        Args:
            client: OpenAI client instance
            batch_texts: Texts to embed
            model: Model name
            normalize: Whether to normalize vectors

        Returns:
            Tuple of (embedding results, token count)
        """
        # Prepare and execute API request
        request_params = self._prepare_request_params(batch_texts, model)
        response: t.Any = await client.embeddings.create(**request_params)

        # Process response data
        results = self._extract_embedding_results(
            response,
            batch_texts,
            normalize,
        )

        # Extract token usage
        tokens = (
            response.usage.total_tokens
            if hasattr(response, "usage") and response.usage
            else 0
        )

        return results, tokens

    def _prepare_request_params(
        self,
        batch_texts: list[str],
        model: str,
    ) -> dict[str, t.Any]:
        """Prepare OpenAI API request parameters.

        Args:
            batch_texts: Texts to embed
            model: Model name

        Returns:
            Request parameters dictionary
        """
        settings = t.cast(OpenAIEmbeddingSettings, self._settings)
        request_params: dict[str, t.Any] = {
            "input": batch_texts,
            "model": model,
            "encoding_format": settings.encoding_format,
        }

        # Add dimensions for v3 models
        if settings.dimensions and model.startswith("text-embedding-3"):
            request_params["dimensions"] = settings.dimensions

        return request_params

    def _extract_embedding_results(
        self,
        response: t.Any,
        batch_texts: list[str],
        normalize: bool,
    ) -> list[EmbeddingResult]:
        """Extract embedding results from API response.

        Args:
            response: OpenAI API response
            batch_texts: Original texts
            normalize: Whether to normalize vectors

        Returns:
            List of embedding results
        """
        results = []

        for i, embedding_data in enumerate(response.data):
            embedding = embedding_data.embedding

            # Normalize if requested
            if normalize:
                embedding = self._normalize_vector(
                    embedding,
                    self._settings.normalization,
                )

            result = EmbeddingResult(
                text=batch_texts[i],
                embedding=embedding,
                model=response.model,
                dimensions=len(embedding),
                tokens=None,  # OpenAI doesn't provide token count per text
                metadata={
                    "index": embedding_data.index,
                    "object": embedding_data.object,
                },
            )
            results.append(result)

        return results

    async def _embed_documents(
        self,
        documents: list[str],
        chunk_size: int,
        chunk_overlap: int,
        model: str,
        **kwargs: t.Any,
    ) -> list[EmbeddingBatch]:
        """Embed large documents with chunking."""
        batches = []

        for document in documents:
            # Split document into chunks
            chunks = self._chunk_text(document, chunk_size, chunk_overlap)

            # Generate embeddings for chunks
            batch = await self._embed_texts(
                chunks,
                model=model,
                normalize=self._settings.normalize_embeddings,
                batch_size=self._settings.batch_size,
                **kwargs,
            )

            # Add document metadata
            for result in batch.results:
                result.metadata.update(
                    {
                        "document_id": hash(document),
                        "is_chunk": True,
                        "chunk_size": chunk_size,
                        "chunk_overlap": chunk_overlap,
                    },
                )

            batches.append(batch)

        return batches

    async def _compute_similarity(
        self,
        embedding1: list[float],
        embedding2: list[float],
        method: str,
    ) -> float:
        """Compute similarity between two embeddings."""
        if method == "cosine":
            return EmbeddingUtils.cosine_similarity(embedding1, embedding2)
        if method == "euclidean":
            return EmbeddingUtils.euclidean_distance(embedding1, embedding2)
        if method == "dot":
            return EmbeddingUtils.dot_product(embedding1, embedding2)
        if method == "manhattan":
            return EmbeddingUtils.manhattan_distance(embedding1, embedding2)
        msg = f"Unsupported similarity method: {method}"
        raise ValueError(msg)

    async def _get_model_info(self, model: str) -> dict[str, t.Any]:
        """Get information about an OpenAI embedding model."""
        model_info: dict[str, t.Any] = {
            "name": model,
            "provider": "openai",
            "type": "embedding",
        }

        # Add model-specific information
        if model == EmbeddingModel.TEXT_EMBEDDING_3_SMALL.value:
            model_info.update(
                {
                    "max_dimensions": 1536,
                    "default_dimensions": 1536,
                    "max_tokens": 8191,
                    "price_per_1k_tokens": 0.00002,
                    "description": "Most efficient embedding model with good performance",
                },
            )
        elif model == EmbeddingModel.TEXT_EMBEDDING_3_LARGE.value:
            model_info.update(
                {
                    "max_dimensions": 3072,
                    "default_dimensions": 3072,
                    "max_tokens": 8191,
                    "price_per_1k_tokens": 0.00013,
                    "description": "Most powerful embedding model with highest accuracy",
                },
            )
        elif model == EmbeddingModel.TEXT_EMBEDDING_ADA_002.value:
            model_info.update(
                {
                    "max_dimensions": 1536,
                    "default_dimensions": 1536,
                    "max_tokens": 8191,
                    "price_per_1k_tokens": 0.0001,
                    "description": "Legacy embedding model (v2)",
                },
            )

        return model_info

    async def _list_models(self) -> list[dict[str, t.Any]]:
        """List available OpenAI embedding models."""
        models = [
            EmbeddingModel.TEXT_EMBEDDING_3_SMALL.value,
            EmbeddingModel.TEXT_EMBEDDING_3_LARGE.value,
            EmbeddingModel.TEXT_EMBEDDING_ADA_002.value,
        ]

        return [await self._get_model_info(model) for model in models]

    async def _apply_rate_limit(self) -> None:
        """Apply rate limiting to API requests."""
        settings = t.cast(OpenAIEmbeddingSettings, self._settings)
        current_time = time.time()
        time_since_last = current_time - self._last_request_time

        # Simple rate limiting - ensure minimum time between requests
        min_interval = 60.0 / settings.requests_per_minute
        if time_since_last < min_interval:
            await asyncio.sleep(min_interval - time_since_last)

        self._last_request_time = time.time()

    async def health_check(self) -> dict[str, t.Any]:
        """Perform health check on OpenAI embeddings service."""
        try:
            # Test with a simple embedding request
            test_result = await self.embed_text("health check test")

            return {
                "status": "healthy",
                "provider": "openai",
                "model": self._settings.model,
                "dimensions": len(test_result),
                "timestamp": datetime.now().isoformat(),
            }
        except Exception as e:
            return {
                "status": "unhealthy",
                "provider": "openai",
                "error": str(e),
                "timestamp": datetime.now().isoformat(),
            }


# Factory function for dependency injection
async def create_openai_embedding(config: Config | None = None) -> OpenAIEmbedding:
    """Create OpenAI embedding adapter instance."""
    if config is None:
        config = await depends.get("config")

    settings = OpenAIEmbeddingSettings()
    return OpenAIEmbedding(settings)


# Type alias
Embedding = OpenAIEmbedding
EmbeddingSettings = OpenAIEmbeddingSettings

depends.set(Embedding, "openai")



"""Security patterns shared between validation actions and services.

This module centralizes security detection patterns to avoid duplication
between the actions layer and services layer.
"""

import re

# Common dangerous patterns for security validation
SQL_INJECTION_PATTERNS: list[re.Pattern[str]] = [
    re.compile(
        r"(\b(union|select|insert|update|delete|drop|create|alter|exec|execute)\b)",
        re.IGNORECASE,
    ),
    re.compile(r"(--|\/\*|\*\/)", re.IGNORECASE),
    # Classic tautology patterns like: ' OR '1'='1' or 1=1
    re.compile(r"\b(?:or|and)\b\s*1\s*=\s*1", re.IGNORECASE),
    re.compile(r"\b(?:or|and)\b\s*'1'\s*=\s*'1'", re.IGNORECASE),
    re.compile(r'\b(?:or|and)\b\s*"1"\s*=\s*"1"', re.IGNORECASE),
    # More generic 'OR x = x' style
    re.compile(r"\b(or|and)\b\s+[^=]+\s*=\s*[^=]+", re.IGNORECASE),
    re.compile(r"(;.*--)", re.IGNORECASE),
    re.compile(r"(\bxp_|\bsp_)", re.IGNORECASE),
]

SCRIPT_INJECTION_PATTERNS: list[re.Pattern[str]] = [
    re.compile(r"<script[^>]*>.*?</script>", re.IGNORECASE | re.DOTALL),
    re.compile(r"javascript:", re.IGNORECASE),
    re.compile(r"vbscript:", re.IGNORECASE),
    re.compile(
        r"on\w+\s*=\s*['\"][^'\"]*['\"]", re.IGNORECASE
    ),  # REGEX OK: XSS detection
    re.compile(r"<iframe[^>]*>.*?</iframe>", re.IGNORECASE | re.DOTALL),
    re.compile(r"<object[^>]*>.*?</object>", re.IGNORECASE | re.DOTALL),
    re.compile(r"<embed[^>]*>.*?</embed>", re.IGNORECASE | re.DOTALL),
]

PATH_TRAVERSAL_PATTERNS: list[re.Pattern[str]] = [
    re.compile(r"\.\./", re.IGNORECASE),  # Parent directory traversal
    re.compile(r"\.\.\\", re.IGNORECASE),  # Windows Parent directory traversal
    re.compile(r"%2e%2e%2f", re.IGNORECASE),  # URL encoded traversal
    re.compile(r"%2e%2e%5c", re.IGNORECASE),  # URL encoded Windows traversal
    re.compile(r"~", re.IGNORECASE),  # Home directory
]


def detect_sql_injection(value: str) -> bool:
    """Check if string contains potential SQL injection patterns.

    Args:
        value: String to check for SQL injection

    Returns:
        True if string appears safe, False if potential injection detected
    """
    value_lower = value.lower()

    for pattern in SQL_INJECTION_PATTERNS:
        if pattern.search(value_lower):
            return False

    return True


def detect_xss(value: str) -> bool:
    """Check if string contains potential XSS/script injection patterns.

    Args:
        value: String to check for XSS

    Returns:
        True if string appears safe, False if potential XSS detected
    """
    value_lower = value.lower()

    for pattern in SCRIPT_INJECTION_PATTERNS:
        if pattern.search(value_lower):
            return False

    return True


def detect_path_traversal(value: str) -> bool:
    """Check if string contains potential path traversal patterns.

    Args:
        value: String to check for path traversal

    Returns:
        True if string appears safe, False if potential traversal detected
    """
    value_lower = value.lower()

    for pattern in PATH_TRAVERSAL_PATTERNS:
        if pattern.search(value_lower):
            return False

    return True



import typing as t
from anyio import Path as AsyncPath
from pydantic import EmailStr, SecretStr

from acb.config import AdapterBase, Config, Settings, gen_password
from acb.depends import Inject, depends


class SmtpBaseSettings(Settings):
    api_key: SecretStr | None = None
    mx_servers: list[str] | None = []
    password: SecretStr = SecretStr(gen_password())
    domain: str | None = None
    port: int = 587
    api_url: str | None = ""
    default_from: EmailStr | None = None
    default_from_name: str | None = None
    test_receiver: EmailStr | None = None
    tls: bool = True
    ssl: bool = False
    template_folder: AsyncPath | None = None
    forwards: dict[str, EmailStr] = {
        "admin": "pat@example.com",
        "info": "terry@example.com",
    }

    @depends.inject
    def __init__(self, config: Inject[Config], **values: t.Any) -> None:
        super().__init__(**values)
        if "domain" not in values and config.app:
            self.domain = f"mail.{config.app.domain}"
        if "default_from" not in values and config.app:
            self.default_from = f"info@{config.app.domain}"
        if "default_from_name" not in values and config.app:
            self.default_from_name = config.app.title


class SmtpBase(AdapterBase): ...



"""Comprehensive ACB Architecture Example.

This example demonstrates how all ACB architectural layers work together:
- Services: For business logic and lifecycle management
- Orchestration: Events, Tasks, and Workflows for communication and process management
- Adapters: For external system integration
- Core: Configuration and dependency injection
"""

import asyncio
from datetime import datetime
from typing import Any

from acb.adapters import import_adapter
from acb.depends import depends
from acb.events import (
    Event,
    EventHandlerResult,
    EventPublisher,
    create_event,
    event_handler,
)
from acb.services._base import ServiceBase, ServiceConfig, ServiceSettings
from acb.tasks import TaskData, task_handler


class UserManagementSettings(ServiceSettings):
    """Settings for user management service."""

    default_user_role: str = "user"
    enable_email_notifications: bool = True
    notification_delay_seconds: float = 1.0


class UserManagementService(ServiceBase):
    """Service for managing users with full lifecycle management."""

    def __init__(self) -> None:
        service_config = ServiceConfig(
            service_id="user_management_service",
            name="User Management Service",
            description="Handles user lifecycle operations",
            dependencies=["cache", "sql"],
            priority=60,
        )
        super().__init__(service_config, UserManagementSettings())

        # Service state
        self._users_created = 0
        self._publisher: EventPublisher | None = None

    async def _initialize(self) -> None:
        """Initialize the service."""
        self.logger.info(f"Initializing {self.name}")

        # Initialize event publisher
        self._publisher = EventPublisher()
        await self._publisher.initialize()

    async def _shutdown(self) -> None:
        """Shutdown the service."""
        if self._publisher:
            await self._publisher.shutdown()

    async def _health_check(self) -> dict[str, Any]:
        """Check service health."""
        # Type the settings to the specific subclass
        settings = self._settings
        return {
            "status": "healthy",
            "users_created": self._users_created,
            "email_notifications_enabled": getattr(
                settings, "enable_email_notifications", True
            ),
        }

    @property
    def users_created(self) -> int:
        """Get count of users created."""
        return self._users_created

    async def create_user(self, user_data: dict[str, Any]) -> dict[str, Any]:
        """Create a new user and emit event."""
        # Import adapters for typing
        Cache = import_adapter("cache")
        SQL = import_adapter("sql")

        # Create user record
        settings = self._settings
        user = {
            "id": user_data.get("id", self._users_created + 1),
            "name": user_data.get("name"),
            "email": user_data.get("email"),
            "role": user_data.get(
                "role", getattr(settings, "default_user_role", "user")
            ),
            "created_at": datetime.now().isoformat(),
        }

        # Save to database and cache - access them through dependency injection
        # For this simplified example, we'll assume they have the expected interface
        try:
            sql: Any = depends.get(SQL)
            if hasattr(sql, "get_session") and callable(sql.get_session):
                # For demonstration purposes only - in real usage, handle properly
                print(f"Saving user to database: {user}")
        except Exception:
            print("Could not access SQL adapter, skipping database save")

        try:
            cache: Any = depends.get(Cache)
            if hasattr(cache, "set") and callable(cache.set):
                await cache.set(f"user:{user['id']}", user, ttl=3600)
            else:
                print(
                    f"Cache doesn't have set method, skipping cache update for user {user['id']}"
                )
        except Exception:
            print(
                f"Could not access cache adapter, skipping cache update for user {user['id']}"
            )

        # Update service metrics
        self._users_created += 1

        # Emit event
        if self._publisher:
            user_event = create_event(
                "user.created",
                "user_management_service",
                {
                    "user_id": user["id"],
                    "email": user["email"],
                    "name": user["name"],
                    "timestamp": user["created_at"],
                },
            )
            await self._publisher.publish(user_event)

        return user


@event_handler("user.created")
async def handle_user_created(event: Event) -> EventHandlerResult:
    """Handle user creation event."""
    user_id = event.payload.get("user_id")
    email = event.payload.get("email")

    print(f"Event handler: Processing new user {user_id} ({email})")

    # In a real app, you might send a welcome email, update statistics, etc.
    # For demo, we'll simulate with a task
    if True:  # Assuming email notifications are enabled
        # Enqueue a task to send welcome email
        from acb.tasks import create_queue

        async with create_queue("memory") as queue:
            email_task = TaskData(
                task_type="send_welcome_email",
                queue_name="default",
                payload={"user_email": email, "user_id": user_id},
            )
            await queue.enqueue(email_task)

    return EventHandlerResult(success=True)


@task_handler("send_welcome_email")
async def send_welcome_email_task(task_data: TaskData) -> dict[str, Any]:
    """Task to send welcome email."""
    payload = task_data.payload
    user_email = payload.get("user_email")
    user_id = payload.get("user_id")

    print(f"Task: Sending welcome email to {user_email} (user {user_id})")

    # Simulate email sending delay
    await asyncio.sleep(1.0)

    return {"status": "email_sent", "user_email": user_email, "user_id": user_id}


async def main() -> None:
    """Demonstrate the complete architecture."""
    print("=== ACB Architecture Example ===")
    print("Demonstrating Services + Events + Tasks + Adapters working together\n")

    # 1. Create and initialize service
    user_service = UserManagementService()
    await user_service.initialize()

    # 2. Create a user (this triggers events and tasks)
    print("1. Creating a new user...")
    new_user = await user_service.create_user(
        {"name": "John Doe", "email": "john@example.com", "role": "user"}
    )
    print(f"   Created user: {new_user}\n")

    # 3. Wait a bit for event processing and task execution
    print("2. Waiting for event handlers and tasks to execute...")
    await asyncio.sleep(2)  # Allow time for event/task processing
    print("   Event handlers and tasks completed\n")

    # 4. Check service health
    print("3. Checking service health...")
    health = await user_service.health_check()
    print(f"   Health status: {health}\n")

    # 5. Show final metrics
    print("4. Service metrics:")
    print(f"   Total users created: {user_service.users_created}")

    # 6. Shutdown service
    await user_service.shutdown()
    print("\nService shutdown completed")

    print("\n=== Architecture Layers Demonstrated ===")
    print("✓ Services: UserManagementService with lifecycle management")
    print("✓ Events: Event publishing and handling for user creation")
    print("✓ Tasks: Background email sending")
    print("✓ Adapters: Cache and SQL integration")
    print("✓ Core: Dependency injection and configuration")


if __name__ == "__main__":
    asyncio.run(main())



"""Tests for aiormq messaging adapter."""

from unittest.mock import AsyncMock, MagicMock, patch
from uuid import uuid4

import pytest

from acb.adapters import import_adapter
from acb.adapters.messaging._base import (
    MessagePriority,
    MessagingConnectionError,
    QueueMessage,
)
from acb.depends import depends


@pytest.fixture
def mock_aiormq_imports():
    """Mock aiormq imports."""
    mock_connection = MagicMock()
    mock_connection.is_closed = False

    mock_channel = MagicMock()
    mock_channel.set_qos = AsyncMock()
    mock_channel.declare_exchange = AsyncMock()
    mock_channel.declare_queue = AsyncMock()
    mock_channel.close = AsyncMock()

    mock_queue = MagicMock()
    mock_queue.bind = AsyncMock()
    mock_queue.declare = AsyncMock(return_value=MagicMock(message_count=0))
    mock_queue.get = AsyncMock(return_value=None)
    mock_queue.consume = AsyncMock(return_value="consumer-tag-123")
    mock_queue.purge = AsyncMock(return_value=0)
    mock_queue.delete = AsyncMock()

    mock_channel.declare_queue.return_value = mock_queue
    mock_connection.channel = AsyncMock(return_value=mock_channel)
    mock_connection.close = AsyncMock()

    mock_exchange = MagicMock()
    mock_exchange.publish = AsyncMock()
    mock_channel.declare_exchange.return_value = mock_exchange

    mock_message = MagicMock()
    mock_message.message_id = str(uuid4())
    mock_message.body = b"test message"
    mock_message.correlation_id = "test-correlation"
    mock_message.headers = {"test": "header"}
    mock_message.ack = AsyncMock()
    mock_message.nack = AsyncMock()

    mock_connect = AsyncMock(return_value=mock_connection)

    # Minimal ExchangeType stub to mimic aiormq.enums.ExchangeType
    class _ExchangeType(str):
        DIRECT = "direct"
        TOPIC = "topic"
        FANOUT = "fanout"
        HEADERS = "headers"

        def __new__(cls, value: str):  # type: ignore[override]
            return str.__new__(cls, value)

    mock_imports = {
        "aiormq": MagicMock(),
        "connect": mock_connect,
        "Message": MagicMock,
        "ExchangeType": _ExchangeType,
    }

    return {
        "imports": mock_imports,
        "connection": mock_connection,
        "channel": mock_channel,
        "queue": mock_queue,
        "exchange": mock_exchange,
        "message": mock_message,
        "connect": mock_connect,
    }


@pytest.fixture
def aiormq_settings():
    """Create aiormq messaging settings."""
    from acb.adapters.messaging.aiormq import AioRmqMessagingSettings

    return AioRmqMessagingSettings(
        connection_url="amqp://guest:guest@localhost:5672/",
        exchange_name="test.exchange",
        queue_durable=True,
        max_priority=10,
        prefetch_count=5,
    )


@pytest.fixture
def aiormq_adapter(aiormq_settings, mock_aiormq_imports):
    """Create aiormq messaging adapter with mocked dependencies."""
    # Register a mock logger to satisfy dependency injection
    mock_logger = MagicMock()
    # Register against the resolved logger adapter class used by depends
    logger_cls = import_adapter("logger")
    depends.set(logger_cls, mock_logger)

    with patch(
        "acb.adapters.messaging.aiormq._get_aiormq_imports",
        return_value=mock_aiormq_imports["imports"],
    ):
        from acb.adapters.messaging.aiormq import AioRmqMessaging

        adapter = AioRmqMessaging(aiormq_settings)

        # Provide lightweight connect/disconnect stubs to avoid event-loop quirks
        async def _stub_connect() -> None:
            # Mark the underlying connect mock as used (parity with real path)
            await mock_aiormq_imports["connect"](
                adapter._settings.connection_url,
                heartbeat=adapter._settings.heartbeat,
                timeout=adapter._settings.connection_timeout,
            )
            adapter._connection = mock_aiormq_imports["connection"]
            adapter._channel = mock_aiormq_imports["channel"]
            adapter._exchange = mock_aiormq_imports["exchange"]
            adapter._connected = True

        async def _stub_disconnect() -> None:
            adapter._connected = False
            await mock_aiormq_imports["channel"].close()

        adapter.connect = AsyncMock(side_effect=_stub_connect)
        adapter.disconnect = AsyncMock(side_effect=_stub_disconnect)

        yield adapter

    # Cleanup
    depends.clear()


@pytest.mark.unit
class TestAioRmqMessagingBasics:
    """Test basic aiormq messaging adapter functionality."""

    def test_adapter_creation(self, aiormq_settings):
        """Test adapter can be created with settings."""
        with patch("acb.adapters.messaging.aiormq._get_aiormq_imports"):
            from acb.adapters.messaging.aiormq import AioRmqMessaging

            adapter = AioRmqMessaging(aiormq_settings)

            assert (
                adapter._settings.connection_url == "amqp://guest:guest@localhost:5672/"
            )
            assert adapter._settings.exchange_name == "test.exchange"
            assert adapter._settings.max_priority == 10
            assert adapter._connection is None
            assert not adapter._connected

    def test_adapter_metadata(self):
        """Test adapter has proper metadata."""
        from acb.adapters.messaging.aiormq import MODULE_METADATA

        assert MODULE_METADATA.name == "aiormq Messaging"
        assert MODULE_METADATA.category == "messaging"
        assert MODULE_METADATA.provider == "rabbitmq"
        assert "aiormq>=6.7.1" in MODULE_METADATA.required_packages

    async def test_connection_lifecycle(self, aiormq_adapter, mock_aiormq_imports):
        """Test connection and disconnection."""
        # Connect
        await aiormq_adapter.connect()

        assert aiormq_adapter._connected
        assert mock_aiormq_imports["connect"].called

        # Disconnect
        await aiormq_adapter.disconnect()

        assert not aiormq_adapter._connected
        assert mock_aiormq_imports["channel"].close.called


@pytest.mark.unit
class TestAioRmqMessagingQueue:
    """Test queue operations."""

    async def test_enqueue_message(self, aiormq_adapter, mock_aiormq_imports):
        """Test enqueueing a message to a queue."""
        await aiormq_adapter.connect()

        message_id = await aiormq_adapter.enqueue(
            "test-queue",
            b"test payload",
            priority=MessagePriority.HIGH,
        )

        assert message_id is not None
        assert mock_aiormq_imports["exchange"].publish.called

    async def test_dequeue_message(self, aiormq_adapter, mock_aiormq_imports):
        """Test dequeuing a message from a queue."""
        await aiormq_adapter.connect()

        # Mock a message return
        mock_message = mock_aiormq_imports["message"]
        mock_aiormq_imports["queue"].get.return_value = mock_message

        message = await aiormq_adapter.dequeue("test-queue")

        assert message is not None
        assert isinstance(message, QueueMessage)
        assert message.payload == b"test message"

    async def test_acknowledge_message(self, aiormq_adapter, mock_aiormq_imports):
        """Test acknowledging a message."""
        await aiormq_adapter.connect()

        # Simulate receiving a message
        mock_message = mock_aiormq_imports["message"]
        mock_aiormq_imports["queue"].get.return_value = mock_message

        message = await aiormq_adapter.dequeue("test-queue")
        assert message is not None

        # Acknowledge it
        await aiormq_adapter.acknowledge("test-queue", str(message.message_id))

        assert mock_message.ack.called

    async def test_reject_message(self, aiormq_adapter, mock_aiormq_imports):
        """Test rejecting a message."""
        await aiormq_adapter.connect()

        # Simulate receiving a message
        mock_message = mock_aiormq_imports["message"]
        mock_aiormq_imports["queue"].get.return_value = mock_message

        message = await aiormq_adapter.dequeue("test-queue")
        assert message is not None

        # Reject with requeue
        await aiormq_adapter.reject("test-queue", str(message.message_id), requeue=True)

        assert mock_message.nack.called

    async def test_enqueue_delayed_ttl_dlx(
        self, aiormq_adapter, mock_aiormq_imports, monkeypatch
    ):
        """Test delayed send path using TTL + DLQ pattern."""
        await aiormq_adapter.connect()

        # Ensure plugin path is disabled so TTL+DLQ is used
        aiormq_adapter._settings.enable_delayed_plugin = False

        # Freeze time for predictable temp queue name
        monkeypatch.setattr(
            "acb.adapters.messaging.aiormq.time.time",
            lambda: 123,
        )

        # Enqueue with delay
        await aiormq_adapter.enqueue("test-delay", b"payload", delay_seconds=0.2)

        # Verify temp queue declared with TTL and DLX arguments
        mock_channel = mock_aiormq_imports["channel"]
        assert mock_channel.declare_queue.await_count >= 1
        args = mock_channel.declare_queue.await_args.kwargs
        assert args["durable"] is False
        assert args["auto_delete"] is True
        arguments = args["arguments"]
        assert arguments["x-message-ttl"] == 200  # 0.2s -> 200ms
        assert (
            arguments["x-dead-letter-exchange"]
            == aiormq_adapter._settings.exchange_name
        )
        assert arguments["x-dead-letter-routing-key"] == "test-delay"

        # Verify publish to temp queue
        temp_name = "delayed.test-delay.123"
        mock_exchange = mock_aiormq_imports["exchange"]
        assert mock_exchange.publish.await_count >= 1
        pub_kwargs = mock_exchange.publish.await_args.kwargs
        assert pub_kwargs["routing_key"] == temp_name


@pytest.mark.unit
class TestAioRmqMessagingPubSub:
    """Test pub/sub operations."""

    async def test_publish_message(self, aiormq_adapter, mock_aiormq_imports):
        """Test publishing a message to a topic."""
        await aiormq_adapter.connect()

        await aiormq_adapter.publish(
            "test.topic",
            b"test payload",
            headers={"test": "header"},
        )

        assert mock_aiormq_imports["exchange"].publish.called

    async def test_subscribe_topic(self, aiormq_adapter, mock_aiormq_imports):
        """Test subscribing to a topic."""
        await aiormq_adapter.connect()

        subscription = await aiormq_adapter.subscribe("test.topic")

        assert subscription is not None
        assert subscription.topic == "test.topic"


@pytest.mark.unit
class TestAioRmqMessagingManagement:
    """Test queue management operations."""

    async def test_purge_queue(self, aiormq_adapter, mock_aiormq_imports):
        """Test purging a queue."""
        await aiormq_adapter.connect()

        mock_aiormq_imports["queue"].purge.return_value = 5

        purged = await aiormq_adapter.purge_queue("test-queue")

        assert purged == 5
        assert mock_aiormq_imports["queue"].purge.called

    async def test_get_queue_stats(self, aiormq_adapter, mock_aiormq_imports):
        """Test getting queue statistics."""
        await aiormq_adapter.connect()

        stats = await aiormq_adapter.get_queue_stats("test-queue")

        assert "message_count" in stats
        assert "consumer_count" in stats


@pytest.mark.unit
class TestAioRmqMessagingCapabilities:
    """Test adapter capabilities."""

    def test_get_capabilities(self, aiormq_adapter):
        """Test getting adapter capabilities."""
        from acb.adapters.messaging._base import MessagingCapability

        capabilities = aiormq_adapter.get_capabilities()

        assert MessagingCapability.BASIC_QUEUE in capabilities
        assert MessagingCapability.PUB_SUB in capabilities
        assert MessagingCapability.PRIORITY_QUEUE in capabilities
        assert MessagingCapability.DELAYED_MESSAGES in capabilities
        assert MessagingCapability.DEAD_LETTER_QUEUE in capabilities


@pytest.mark.unit
class TestAioRmqMessagingHealth:
    """Test health check functionality."""

    async def test_health_check_healthy(self, aiormq_adapter):
        """Test health check when connection is healthy."""
        await aiormq_adapter.connect()

        health = await aiormq_adapter.health_check()

        assert health["healthy"]
        assert health["connected"]
        assert "latency_ms" in health


@pytest.mark.unit
class TestAioRmqMessagingErrors:
    """Test error handling."""

    async def test_operation_without_connection(self, aiormq_settings):
        """Test operations fail without connection."""
        with patch("acb.adapters.messaging.aiormq._get_aiormq_imports"):
            from acb.adapters.messaging.aiormq import AioRmqMessaging

            adapter = AioRmqMessaging(aiormq_settings)

            with pytest.raises(MessagingConnectionError):
                await adapter.enqueue("test-queue", b"test")



"""Repository Base Classes and Interface.

Provides foundational repository pattern implementation with:
- Generic repository interface for CRUD operations
- Base settings and configuration
- Error handling for repository operations
- Integration with ACB dependency injection
"""

import builtins
from abc import ABC, abstractmethod
from enum import Enum

import typing as t
from dataclasses import dataclass
from pydantic import Field, field_validator
from typing import Any, TypeVar

from acb.cleanup import CleanupMixin
from acb.config import Settings
from acb.depends import depends

# Type variables for generic repository
EntityType = TypeVar("EntityType")
IDType = TypeVar("IDType")


class RepositoryError(Exception):
    """Base exception for repository operations."""

    def __init__(
        self,
        message: str,
        entity_type: str | None = None,
        operation: str | None = None,
    ) -> None:
        self.entity_type = entity_type
        self.operation = operation
        super().__init__(message)


class EntityNotFoundError(RepositoryError):
    """Raised when an entity is not found."""

    def __init__(self, entity_type: str, entity_id: Any) -> None:
        super().__init__(
            f"{entity_type} with ID {entity_id} not found",
            entity_type=entity_type,
            operation="find",
        )
        self.entity_id = entity_id


class DuplicateEntityError(RepositoryError):
    """Raised when trying to create a duplicate entity."""

    def __init__(self, entity_type: str, conflict_field: str, value: Any) -> None:
        super().__init__(
            f"{entity_type} with {conflict_field}={value} already exists",
            entity_type=entity_type,
            operation="create",
        )
        self.conflict_field = conflict_field
        self.value = value


class SortDirection(Enum):
    """Sort direction enumeration."""

    ASC = "asc"
    DESC = "desc"


@dataclass
class SortCriteria:
    """Sort criteria specification."""

    field: str
    direction: SortDirection = SortDirection.ASC


@dataclass
class PaginationInfo:
    """Pagination information."""

    page: int = 1
    page_size: int = 50
    total_items: int | None = None
    total_pages: int | None = None

    def __post_init__(self) -> None:
        if self.total_items is not None and self.total_pages is None:
            self.total_pages = (self.total_items + self.page_size - 1) // self.page_size

    @property
    def offset(self) -> int:
        """Calculate offset for database queries."""
        return (self.page - 1) * self.page_size

    @property
    def has_next(self) -> bool:
        """Check if there's a next page."""
        return self.total_pages is not None and self.page < self.total_pages

    @property
    def has_previous(self) -> bool:
        """Check if there's a previous page."""
        return self.page > 1


class RepositorySettings(Settings):
    """Repository configuration settings."""

    # Caching settings
    cache_enabled: bool = True
    cache_ttl: int = Field(default=300, description="Cache TTL in seconds")
    cache_prefix: str = Field(default="repo", description="Cache key prefix")

    # Query settings
    default_page_size: int = Field(default=50, ge=1, le=1000)
    max_page_size: int = Field(default=1000, ge=1)
    query_timeout: float = Field(default=30.0, description="Query timeout in seconds")

    # Transaction settings
    transaction_timeout: float = Field(
        default=60.0,
        description="Transaction timeout in seconds",
    )
    isolation_level: str = Field(
        default="READ_COMMITTED",
        description="Default isolation level",
    )

    # Performance settings
    batch_size: int = Field(default=100, ge=1, le=10000)
    connection_pool_size: int = Field(default=10, ge=1, le=100)

    @field_validator("default_page_size")
    @classmethod
    def validate_page_size(cls, v: int, info: t.Any) -> int:
        values: Any = info.data if hasattr(info, "data") else {}
        if "max_page_size" in values and v > values["max_page_size"]:
            msg = "default_page_size cannot exceed max_page_size"
            raise ValueError(msg)
        return v


class RepositoryProtocol(t.Protocol[EntityType, IDType]):  # type: ignore[misc]
    """Protocol defining repository interface."""

    async def create(self, entity: EntityType) -> EntityType:
        """Create a new entity."""
        ...

    async def get_by_id(self, entity_id: IDType) -> EntityType | None:
        """Get entity by ID."""
        ...

    async def update(self, entity: EntityType) -> EntityType:
        """Update an existing entity."""
        ...

    async def delete(self, entity_id: IDType) -> bool:
        """Delete entity by ID."""
        ...

    async def list(
        self,
        filters: dict[str, Any] | None = None,
        sort: list[SortCriteria] | None = None,
        pagination: PaginationInfo | None = None,
    ) -> list[EntityType]:
        """List entities with filtering, sorting, and pagination."""
        ...

    async def count(self, filters: dict[str, Any] | None = None) -> int:
        """Count entities matching filters."""
        ...

    async def exists(self, entity_id: IDType) -> bool:
        """Check if entity exists."""
        ...


class RepositoryBase[EntityType, IDType](CleanupMixin, ABC):
    """Abstract base class for repositories.

    Provides common functionality for all repository implementations:
    - CRUD operations interface
    - Error handling and logging
    - Caching integration
    - Performance monitoring
    - Resource cleanup
    """

    def __init__(
        self,
        entity_type: type[EntityType],
        settings: RepositorySettings | None = None,
    ) -> None:
        super().__init__()
        self.entity_type = entity_type
        self.entity_name = getattr(entity_type, "__name__", str(entity_type))
        self.settings = settings or depends.get_sync(RepositorySettings)
        self._cache = None
        self._metrics: dict[str, t.Any] = {}

    async def _async_init(self) -> None:
        """Async initialization of repository dependencies."""
        if self._cache is None:
            from acb.adapters import import_adapter

            Cache = import_adapter("cache")
            self._cache = await depends.get(Cache)

    @property
    def cache_key_prefix(self) -> str:
        """Get cache key prefix for this repository."""
        return f"{self.settings.cache_prefix}:{self.entity_name.lower()}"

    def _build_cache_key(self, key: str) -> str:
        """Build cache key with proper prefix."""
        return f"{self.cache_key_prefix}:{key}"

    async def _get_cache(self) -> None:
        """Get cache adapter if enabled."""
        if self.settings.cache_enabled and self._cache is None:
            try:
                from acb.adapters import import_adapter

                Cache = import_adapter("cache")
                self._cache = await depends.get(Cache)
            except ImportError:
                self.settings.cache_enabled = False
        return self._cache

    async def _increment_metric(self, operation: str, success: bool = True) -> None:
        """Track operation metrics."""
        metric_key = f"{operation}_{'success' if success else 'error'}"
        self._metrics[metric_key] = self._metrics.get(metric_key, 0) + 1

    async def _handle_error(self, error: Exception, operation: str) -> None:
        """Handle repository errors with proper logging and metrics."""
        await self._increment_metric(operation, success=False)

        if isinstance(error, RepositoryError):
            raise

        # Wrap other exceptions in RepositoryError
        msg = f"Repository operation failed: {error}"
        raise RepositoryError(
            msg,
            entity_type=self.entity_name,
            operation=operation,
        ) from error

    @abstractmethod
    async def create(self, entity: EntityType) -> EntityType:
        """Create a new entity.

        Args:
            entity: Entity to create

        Returns:
            Created entity with any generated fields

        Raises:
            DuplicateEntityError: If entity already exists
            RepositoryError: For other creation failures
        """

    @abstractmethod
    async def get_by_id(self, entity_id: IDType) -> EntityType | None:
        """Get entity by ID.

        Args:
            entity_id: Unique identifier for the entity

        Returns:
            Entity if found, None otherwise
        """

    async def get_by_id_or_raise(self, entity_id: IDType) -> EntityType:
        """Get entity by ID, raise if not found.

        Args:
            entity_id: Unique identifier for the entity

        Returns:
            Entity if found

        Raises:
            EntityNotFoundError: If entity not found
        """
        entity = await self.get_by_id(entity_id)
        if entity is None:
            raise EntityNotFoundError(self.entity_name, entity_id)
        return entity

    @abstractmethod
    async def update(self, entity: EntityType) -> EntityType:
        """Update an existing entity.

        Args:
            entity: Entity to update

        Returns:
            Updated entity

        Raises:
            EntityNotFoundError: If entity doesn't exist
            RepositoryError: For other update failures
        """

    @abstractmethod
    async def delete(self, entity_id: IDType) -> bool:
        """Delete entity by ID.

        Args:
            entity_id: Unique identifier for the entity

        Returns:
            True if entity was deleted, False if not found
        """

    async def delete_or_raise(self, entity_id: IDType) -> None:
        """Delete entity by ID, raise if not found.

        Args:
            entity_id: Unique identifier for the entity

        Raises:
            EntityNotFoundError: If entity not found
        """
        deleted = await self.delete(entity_id)
        if not deleted:
            raise EntityNotFoundError(self.entity_name, entity_id)

    @abstractmethod
    async def list(
        self,
        filters: dict[str, Any] | None = None,
        sort: list[SortCriteria] | None = None,
        pagination: PaginationInfo | None = None,
    ) -> list[EntityType]:
        """List entities with filtering, sorting, and pagination.

        Args:
            filters: Dictionary of field filters
            sort: List of sort criteria
            pagination: Pagination information

        Returns:
            List of entities matching criteria
        """

    @abstractmethod
    async def count(self, filters: dict[str, Any] | None = None) -> int:
        """Count entities matching filters.

        Args:
            filters: Dictionary of field filters

        Returns:
            Number of matching entities
        """

    async def exists(self, entity_id: IDType) -> bool:
        """Check if entity exists.

        Args:
            entity_id: Unique identifier for the entity

        Returns:
            True if entity exists, False otherwise
        """
        entity = await self.get_by_id(entity_id)
        return entity is not None

    async def list_paginated(
        self,
        filters: dict[str, Any] | None = None,
        sort: builtins.list[SortCriteria] | None = None,
        page: int = 1,
        page_size: int | None = None,
    ) -> tuple[builtins.list[EntityType], PaginationInfo]:
        """List entities with pagination information.

        Args:
            filters: Dictionary of field filters
            sort: List of sort criteria
            page: Page number (1-based)
            page_size: Items per page

        Returns:
            Tuple of (entities, pagination_info)
        """
        if page_size is None:
            page_size = self.settings.default_page_size

        page_size = min(page_size, self.settings.max_page_size)

        # Get total count
        total_items = await self.count(filters)

        # Create pagination info
        pagination = PaginationInfo(
            page=page,
            page_size=page_size,
            total_items=total_items,
        )

        # Get entities for current page
        entities = await self.list(filters, sort, pagination)

        return entities, pagination

    async def batch_create(
        self,
        entities: builtins.list[EntityType],
    ) -> builtins.list[EntityType]:
        """Create multiple entities in batch.

        Args:
            entities: List of entities to create

        Returns:
            List of created entities
        """
        return [await self.create(entity) for entity in entities]

    async def batch_update(
        self,
        entities: builtins.list[EntityType],
    ) -> builtins.list[EntityType]:
        """Update multiple entities in batch.

        Args:
            entities: List of entities to update

        Returns:
            List of updated entities
        """
        return [await self.update(entity) for entity in entities]

    async def batch_delete(self, entity_ids: builtins.list[IDType]) -> int:
        """Delete multiple entities by ID.

        Args:
            entity_ids: List of entity IDs to delete

        Returns:
            Number of entities deleted
        """
        deleted_count = 0
        for entity_id in entity_ids:
            if await self.delete(entity_id):
                deleted_count += 1
        return deleted_count

    async def get_metrics(self) -> dict[str, Any]:
        """Get repository performance metrics.

        Returns:
            Dictionary of metrics
        """
        return {
            "entity_type": self.entity_name,
            "cache_enabled": self.settings.cache_enabled,
            "operations": self._metrics.copy(),
            "settings": {
                "default_page_size": self.settings.default_page_size,
                "max_page_size": self.settings.max_page_size,
                "cache_ttl": self.settings.cache_ttl,
            },
        }

    async def _cleanup_resources(self) -> None:
        """Clean up repository resources."""
        self._metrics.clear()
        self._cache = None



"""Tests for Custom reasoning adapter."""

from unittest.mock import MagicMock

import pytest

from acb.adapters.reasoning._base import (
    DecisionTree,
    MemoryType,
    ReasoningContext,
    ReasoningRequest,
    ReasoningStep,
    ReasoningStrategy,
)
from acb.adapters.reasoning.custom import EnhancedRule, Reasoning, RuleEngine

DecisionNode = DecisionTree  # Alias for backward compatibility


class MockCustomSettings:
    """Mock settings for Custom reasoning adapter."""

    def __init__(self):
        self.max_reasoning_depth = 10
        self.enable_parallel_reasoning = True
        self.confidence_threshold = 0.6
        self.enable_self_reflection = True
        self.enable_step_validation = True


@pytest.fixture
def mock_settings():
    """Mock adapter settings."""
    return MockCustomSettings()


@pytest.fixture
def mock_config(mock_settings):
    """Mock config with reasoning settings."""
    config = MagicMock()
    config.reasoning = mock_settings
    return config


@pytest.fixture
def reasoning_adapter(mock_config):
    """Create reasoning adapter with mocked dependencies."""
    adapter = Reasoning()
    adapter._settings = mock_config.reasoning
    return adapter


@pytest.fixture
def sample_rule():
    """Create a sample rule for testing."""
    return EnhancedRule(
        id="test_rule",
        name="Test Rule",
        description="A test rule",
        conditions=[
            {"field": "age", "operator": ">=", "value": 18},
            {"field": "status", "operator": "==", "value": "active"},
        ],
        action="approve",
        priority=1,
        confidence=0.9,
        tags=["test", "approval"],
    )


@pytest.fixture
def sample_decision_tree():
    """Create a sample decision tree for testing."""
    return DecisionTree(
        id="test_tree",
        name="Test Decision Tree",
        description="A test decision tree",
        root_node=DecisionNode(
            id="root",
            condition={"field": "age", "operator": ">=", "value": 18},
            true_branch=DecisionNode(
                id="adult",
                condition={"field": "income", "operator": ">", "value": 50000},
                true_branch=DecisionNode(
                    id="high_income", action="premium_approve", confidence=0.9
                ),
                false_branch=DecisionNode(
                    id="low_income", action="standard_approve", confidence=0.7
                ),
            ),
            false_branch=DecisionNode(id="minor", action="reject", confidence=0.95),
        ),
    )


class TestCustomSettings:
    """Test settings validation and initialization."""

    def test_settings_initialization(self, mock_settings):
        """Test settings are properly initialized."""
        assert mock_settings.max_reasoning_depth == 10
        assert mock_settings.enable_parallel_reasoning is True
        assert mock_settings.confidence_threshold == 0.6
        assert mock_settings.enable_self_reflection is True
        assert mock_settings.enable_step_validation is True


class TestRuleEngine:
    """Test rule engine functionality."""

    def test_rule_engine_initialization(self):
        """Test rule engine initialization."""
        engine = RuleEngine()
        assert engine is not None
        assert len(engine.rules) == 0

    def test_add_rule(self, sample_rule):
        """Test adding rules to engine."""
        engine = RuleEngine()
        engine.add_rule(sample_rule)

        assert len(engine.rules) == 1
        assert engine.rules[0] == sample_rule

    def test_evaluate_rule_success(self, sample_rule):
        """Test successful rule evaluation."""
        engine = RuleEngine()
        data = {"age": 25, "status": "active"}

        result = engine.evaluate_rule(sample_rule, data)

        assert result.rule_id == "test_rule"
        assert result.matched is True
        assert result.action == "approve"
        assert result.confidence == 0.9

    def test_evaluate_rule_failure(self, sample_rule):
        """Test failed rule evaluation."""
        engine = RuleEngine()
        data = {"age": 16, "status": "active"}

        result = engine.evaluate_rule(sample_rule, data)

        assert result.rule_id == "test_rule"
        assert result.matched is False
        assert result.action is None
        assert result.confidence == 0.0

    def test_evaluate_rule_partial_match(self, sample_rule):
        """Test partial rule evaluation."""
        engine = RuleEngine()
        data = {"age": 25, "status": "inactive"}

        result = engine.evaluate_rule(sample_rule, data)

        assert result.rule_id == "test_rule"
        assert result.matched is False

    def test_evaluate_all_rules(self, sample_rule):
        """Test evaluating all rules."""
        engine = RuleEngine()

        # Add multiple rules
        rule2 = EnhancedRule(
            id="backup_rule",
            name="Backup Rule",
            description="Backup approval rule",
            conditions=[{"field": "score", "operator": ">", "value": 80}],
            action="backup_approve",
            priority=2,
            confidence=0.8,
        )

        engine.add_rule(sample_rule)
        engine.add_rule(rule2)

        data = {"age": 25, "status": "active", "score": 85}

        results = engine.evaluate_all_rules(data)

        assert len(results) == 2
        assert all(result.matched for result in results)
        # Should be sorted by priority
        assert results[0].rule_id == "test_rule"  # priority 1
        assert results[1].rule_id == "backup_rule"  # priority 2

    def test_condition_operators(self):
        """Test various condition operators."""
        engine = RuleEngine()

        # Test different operators
        test_cases = [
            ({"field": "value", "operator": "==", "value": 10}, {"value": 10}, True),
            ({"field": "value", "operator": "!=", "value": 10}, {"value": 5}, True),
            ({"field": "value", "operator": ">", "value": 10}, {"value": 15}, True),
            ({"field": "value", "operator": "<", "value": 10}, {"value": 5}, True),
            ({"field": "value", "operator": ">=", "value": 10}, {"value": 10}, True),
            ({"field": "value", "operator": "<=", "value": 10}, {"value": 10}, True),
            (
                {"field": "name", "operator": "in", "value": ["Alice", "Bob"]},
                {"name": "Alice"},
                True,
            ),
            (
                {"field": "tags", "operator": "contains", "value": "test"},
                {"tags": ["test", "other"]},
                True,
            ),
        ]

        for condition, data, expected in test_cases:
            rule = EnhancedRule(
                id="test",
                name="Test",
                description="Test",
                conditions=[condition],
                action="test",
                priority=1,
                confidence=0.9,
            )

            result = engine.evaluate_rule(rule, data)
            assert result.matched == expected, (
                f"Failed for condition {condition} with data {data}"
            )


class TestDecisionTree:
    """Test decision tree functionality."""

    async def test_decision_tree_evaluation(
        self, reasoning_adapter, sample_decision_tree
    ):
        """Test decision tree evaluation."""
        reasoning_adapter._decision_trees["test_tree"] = sample_decision_tree

        # Test high income adult
        data = {"age": 25, "income": 75000}
        result = await reasoning_adapter._evaluate_decision_tree("test_tree", data)

        assert result.action == "premium_approve"
        assert result.confidence == 0.9
        assert "high_income" in result.path

    async def test_decision_tree_low_income_path(
        self, reasoning_adapter, sample_decision_tree
    ):
        """Test decision tree low income path."""
        reasoning_adapter._decision_trees["test_tree"] = sample_decision_tree

        data = {"age": 30, "income": 30000}
        result = await reasoning_adapter._evaluate_decision_tree("test_tree", data)

        assert result.action == "standard_approve"
        assert result.confidence == 0.7
        assert "low_income" in result.path

    async def test_decision_tree_minor_path(
        self, reasoning_adapter, sample_decision_tree
    ):
        """Test decision tree minor path."""
        reasoning_adapter._decision_trees["test_tree"] = sample_decision_tree

        data = {"age": 16, "income": 0}
        result = await reasoning_adapter._evaluate_decision_tree("test_tree", data)

        assert result.action == "reject"
        assert result.confidence == 0.95
        assert "minor" in result.path

    async def test_decision_tree_missing_data(
        self, reasoning_adapter, sample_decision_tree
    ):
        """Test decision tree with missing data."""
        reasoning_adapter._decision_trees["test_tree"] = sample_decision_tree

        data = {"income": 50000}  # Missing age

        with pytest.raises(ValueError, match="Missing field 'age' in data"):
            await reasoning_adapter._evaluate_decision_tree("test_tree", data)

    async def test_decision_tree_not_found(self, reasoning_adapter):
        """Test decision tree not found."""
        with pytest.raises(ValueError, match="Decision tree 'nonexistent' not found"):
            await reasoning_adapter._evaluate_decision_tree("nonexistent", {})


class TestReasoningOperations:
    """Test core reasoning operations."""

    async def test_rule_based_reasoning(self, reasoning_adapter, sample_rule):
        """Test rule-based reasoning strategy."""
        reasoning_adapter._rule_engine.add_rule(sample_rule)

        request = ReasoningRequest(
            query="Should we approve this user?",
            strategy=ReasoningStrategy.RULE_BASED,
            context=ReasoningContext(data={"age": 25, "status": "active"}),
        )

        response = await reasoning_adapter.reason(request)

        assert response.result == "approve"
        assert response.confidence >= 0.9
        assert response.strategy == ReasoningStrategy.RULE_BASED

    async def test_decision_tree_reasoning(
        self, reasoning_adapter, sample_decision_tree
    ):
        """Test decision tree reasoning strategy."""
        reasoning_adapter._decision_trees["test_tree"] = sample_decision_tree

        request = ReasoningRequest(
            query="What should we do with this application?",
            strategy=ReasoningStrategy.DECISION_TREE,
            context=ReasoningContext(
                decision_tree_id="test_tree", data={"age": 25, "income": 75000}
            ),
        )

        response = await reasoning_adapter.reason(request)

        assert response.result == "premium_approve"
        assert response.confidence == 0.9
        assert response.strategy == ReasoningStrategy.DECISION_TREE

    async def test_step_by_step_reasoning(self, reasoning_adapter):
        """Test step-by-step reasoning."""
        steps = [
            ReasoningStep(
                step_number=1,
                description="Analyze the problem",
                reasoning="First, we need to understand what we're solving",
                result="Problem identified",
                confidence=0.8,
            ),
            ReasoningStep(
                step_number=2,
                description="Generate solution",
                reasoning="Based on the analysis, we can propose a solution",
                result="Solution proposed",
                confidence=0.9,
            ),
        ]

        response = await reasoning_adapter._perform_step_by_step_reasoning(
            "Solve this problem", steps
        )

        assert len(response.steps) == 2
        assert response.result == "Solution proposed"
        assert response.confidence > 0.8

    async def test_parallel_reasoning_paths(self, reasoning_adapter, sample_rule):
        """Test parallel reasoning paths."""
        # Add multiple rules
        rule2 = EnhancedRule(
            id="rule2",
            name="Alternative Rule",
            description="Alternative approval rule",
            conditions=[{"field": "score", "operator": ">", "value": 80}],
            action="score_approve",
            priority=2,
            confidence=0.8,
        )

        reasoning_adapter._rule_engine.add_rule(sample_rule)
        reasoning_adapter._rule_engine.add_rule(rule2)

        paths = [
            {"age": 25, "status": "active"},
            {"score": 85},
            {"age": 30, "status": "active", "score": 90},
        ]

        results = await reasoning_adapter._evaluate_parallel_paths(paths)

        assert len(results) == 3
        assert results[0]["matched_rules"] >= 1
        assert results[2]["matched_rules"] >= 2  # Should match both rules


class TestMemoryOperations:
    """Test memory management operations."""

    async def test_store_memory_episodic(self, reasoning_adapter):
        """Test storing episodic memory."""
        await reasoning_adapter.store_memory(
            "session_1", "User approved loan application", MemoryType.EPISODIC
        )

        memories = reasoning_adapter._memory.get("session_1", {})
        assert MemoryType.EPISODIC in memories
        assert "User approved loan application" in memories[MemoryType.EPISODIC]

    async def test_store_memory_semantic(self, reasoning_adapter):
        """Test storing semantic memory."""
        await reasoning_adapter.store_memory(
            "global", "High income threshold is 50000", MemoryType.SEMANTIC
        )

        memories = reasoning_adapter._memory.get("global", {})
        assert MemoryType.SEMANTIC in memories
        assert "High income threshold is 50000" in memories[MemoryType.SEMANTIC]

    async def test_retrieve_memory_by_type(self, reasoning_adapter):
        """Test retrieving memory by type."""
        reasoning_adapter._memory["session_1"] = {
            MemoryType.EPISODIC: ["Episode 1", "Episode 2"],
            MemoryType.SEMANTIC: ["Fact 1"],
        }

        episodic = await reasoning_adapter.retrieve_memory(
            "session_1", MemoryType.EPISODIC
        )
        semantic = await reasoning_adapter.retrieve_memory(
            "session_1", MemoryType.SEMANTIC
        )

        assert len(episodic) == 2
        assert len(semantic) == 1
        assert "Episode 1" in episodic
        assert "Fact 1" in semantic

    async def test_memory_based_reasoning(self, reasoning_adapter):
        """Test reasoning with memory context."""
        # Store relevant memories
        await reasoning_adapter.store_memory(
            "session_1", "User previously rejected for low income", MemoryType.EPISODIC
        )
        await reasoning_adapter.store_memory(
            "session_1", "Minimum income requirement: 40000", MemoryType.SEMANTIC
        )

        memories = await reasoning_adapter.retrieve_memory("session_1")
        context = ReasoningContext(
            session_id="session_1",
            previous_memories=memories,
        )

        # This would be used in actual reasoning
        assert len(context.previous_memories) == 2


class TestSelfReflection:
    """Test self-reflection capabilities."""

    async def test_reflection_on_decision(self, reasoning_adapter):
        """Test self-reflection on a decision."""
        decision = "approve"
        confidence = 0.7
        reasoning_steps = [
            ReasoningStep(
                step_number=1,
                description="Check age",
                reasoning="User is 25 years old",
                result="Age requirement met",
                confidence=0.9,
            ),
            ReasoningStep(
                step_number=2,
                description="Check status",
                reasoning="User status is active",
                result="Status requirement met",
                confidence=0.8,
            ),
        ]

        reflection = await reasoning_adapter._reflect_on_decision(
            decision, confidence, reasoning_steps
        )

        assert reflection is not None
        assert "confidence" in reflection
        assert "potential_issues" in reflection
        assert "improvements" in reflection

    async def test_confidence_calculation(self, reasoning_adapter):
        """Test confidence calculation from steps."""
        steps = [
            ReasoningStep(
                step_number=1, description="", reasoning="", result="", confidence=0.9
            ),
            ReasoningStep(
                step_number=2, description="", reasoning="", result="", confidence=0.8
            ),
            ReasoningStep(
                step_number=3, description="", reasoning="", result="", confidence=0.7
            ),
        ]

        confidence = reasoning_adapter._calculate_overall_confidence(steps)

        # Should be weighted average or similar calculation
        assert 0.7 <= confidence <= 0.9

    async def test_identify_weak_steps(self, reasoning_adapter):
        """Test identification of weak reasoning steps."""
        steps = [
            ReasoningStep(
                step_number=1,
                description="Strong step",
                reasoning="",
                result="",
                confidence=0.9,
            ),
            ReasoningStep(
                step_number=2,
                description="Weak step",
                reasoning="",
                result="",
                confidence=0.4,
            ),
            ReasoningStep(
                step_number=3,
                description="Medium step",
                reasoning="",
                result="",
                confidence=0.7,
            ),
        ]

        weak_steps = reasoning_adapter._identify_weak_steps(steps, threshold=0.6)

        assert len(weak_steps) == 1
        assert weak_steps[0].description == "Weak step"


class TestToolIntegration:
    """Test tool integration capabilities."""

    async def test_register_decision_tool(self, reasoning_adapter):
        """Test registering a decision tool."""

        def approval_tool(data: dict) -> dict:
            return {"approved": data.get("score", 0) > 70}

        reasoning_adapter.register_tool("approval_checker", approval_tool)

        assert "approval_checker" in reasoning_adapter._tools
        assert reasoning_adapter._tools["approval_checker"] == approval_tool

    async def test_execute_tool(self, reasoning_adapter):
        """Test executing a registered tool."""

        def simple_tool(data: dict) -> dict:
            return {"result": data.get("value", 0) * 2}

        reasoning_adapter.register_tool("doubler", simple_tool)

        result = await reasoning_adapter._execute_tool("doubler", {"value": 5})

        assert result == {"result": 10}

    async def test_execute_nonexistent_tool(self, reasoning_adapter):
        """Test executing a nonexistent tool."""
        with pytest.raises(ValueError, match="Tool 'nonexistent' not found"):
            await reasoning_adapter._execute_tool("nonexistent", {})


class TestErrorHandling:
    """Test error handling scenarios."""

    async def test_invalid_reasoning_strategy(self, reasoning_adapter):
        """Test invalid reasoning strategy."""
        request = ReasoningRequest(
            query="Test query",
            strategy="invalid_strategy",  # type: ignore
        )

        with pytest.raises(ValueError, match="Unsupported reasoning strategy"):
            await reasoning_adapter.reason(request)

    async def test_rule_based_without_data(self, reasoning_adapter):
        """Test rule-based reasoning without data."""
        request = ReasoningRequest(
            query="Test query",
            strategy=ReasoningStrategy.RULE_BASED,
        )

        with pytest.raises(ValueError, match="Data must be provided"):
            await reasoning_adapter.reason(request)

    async def test_decision_tree_without_id(self, reasoning_adapter):
        """Test decision tree reasoning without tree ID."""
        request = ReasoningRequest(
            query="Test query",
            strategy=ReasoningStrategy.DECISION_TREE,
            context=ReasoningContext(data={"age": 25}),
        )

        with pytest.raises(ValueError, match="Decision tree ID must be specified"):
            await reasoning_adapter.reason(request)

    async def test_malformed_rule_condition(self):
        """Test malformed rule condition."""
        engine = RuleEngine()

        # Missing operator
        rule = EnhancedRule(
            id="bad_rule",
            name="Bad Rule",
            description="Bad rule",
            conditions=[{"field": "age", "value": 18}],  # Missing operator
            action="test",
            priority=1,
            confidence=0.9,
        )

        with pytest.raises(KeyError):
            engine.evaluate_rule(rule, {"age": 25})


class TestIntegration:
    """Test integration scenarios."""

    async def test_complex_rule_chain(self, reasoning_adapter):
        """Test complex rule chain evaluation."""
        # Create multiple related rules
        rules = [
            EnhancedRule(
                id="age_check",
                name="Age Check",
                description="Check minimum age",
                conditions=[{"field": "age", "operator": ">=", "value": 18}],
                action="age_approved",
                priority=1,
                confidence=0.9,
            ),
            EnhancedRule(
                id="income_check",
                name="Income Check",
                description="Check minimum income",
                conditions=[
                    {"field": "age", "operator": ">=", "value": 18},
                    {"field": "income", "operator": ">=", "value": 30000},
                ],
                action="income_approved",
                priority=2,
                confidence=0.8,
            ),
            EnhancedRule(
                id="premium_check",
                name="Premium Check",
                description="Check premium eligibility",
                conditions=[
                    {"field": "age", "operator": ">=", "value": 25},
                    {"field": "income", "operator": ">=", "value": 75000},
                    {"field": "credit_score", "operator": ">=", "value": 750},
                ],
                action="premium_approved",
                priority=3,
                confidence=0.95,
            ),
        ]

        for rule in rules:
            reasoning_adapter._rule_engine.add_rule(rule)

        # Test premium candidate
        request = ReasoningRequest(
            query="Evaluate premium application",
            strategy=ReasoningStrategy.RULE_BASED,
            context=ReasoningContext(
                data={"age": 30, "income": 80000, "credit_score": 780}
            ),
        )

        response = await reasoning_adapter.reason(request)

        # Should get the highest priority match
        assert response.result == "premium_approved"
        assert response.confidence >= 0.9

    async def test_memory_informed_reasoning(self, reasoning_adapter, sample_rule):
        """Test reasoning informed by previous memories."""
        # Store relevant history
        await reasoning_adapter.store_memory(
            "user_123",
            "Previous application was rejected for insufficient income",
            MemoryType.EPISODIC,
        )

        reasoning_adapter._rule_engine.add_rule(sample_rule)

        # Retrieve memories for context
        memories = await reasoning_adapter.retrieve_memory("user_123")

        request = ReasoningRequest(
            query="Should we approve this new application?",
            strategy=ReasoningStrategy.RULE_BASED,
            context=ReasoningContext(
                session_id="user_123",
                data={"age": 25, "status": "active"},
                previous_memories=memories,
            ),
        )

        response = await reasoning_adapter.reason(request)

        # Reasoning should succeed but context includes previous rejection
        assert response.result == "approve"
        assert len(response.context.previous_memories) == 1


class TestCleanup:
    """Test resource cleanup."""

    async def test_cleanup_resources(self, reasoning_adapter, sample_rule):
        """Test cleanup of adapter resources."""
        # Setup resources
        reasoning_adapter._rule_engine.add_rule(sample_rule)
        reasoning_adapter._tools["test_tool"] = lambda x: x
        reasoning_adapter._memory["session1"] = {MemoryType.CONVERSATION: ["test"]}

        # Perform cleanup
        await reasoning_adapter.cleanup()

        # Verify resources are cleared
        assert len(reasoning_adapter._rule_engine.rules) == 0
        assert len(reasoning_adapter._tools) == 0
        assert len(reasoning_adapter._memory) == 0
        assert len(reasoning_adapter._decision_trees) == 0



"""Tests for the SMTP base components."""

from unittest.mock import MagicMock, patch

import pytest
from pydantic import SecretStr

from acb.adapters.smtp._base import SmtpBase, SmtpBaseSettings


class MockSmtpBase(SmtpBase):
    def __init__(self) -> None:
        self.config = MagicMock()  # type: ignore
        self.logger = MagicMock()
        self.initialized: bool = False

    async def init(self) -> None:
        self.initialized = True


class TestSmtpBaseSettings:
    def test_init_basic(self) -> None:
        settings: SmtpBaseSettings = SmtpBaseSettings(
            domain="example.com",
            default_from="test@example.com",
            api_key=SecretStr("test_api_key"),
        )
        assert settings.domain == "example.com"
        assert settings.default_from == "test@example.com"
        assert settings.api_key.get_secret_value() == "test_api_key"

    def test_init_defaults(self) -> None:
        settings: SmtpBaseSettings = SmtpBaseSettings(
            domain="example.com",
            default_from="test@example.com",
        )
        assert settings.domain == "example.com"
        assert settings.default_from == "test@example.com"
        assert settings.api_key is None

    def test_init_with_config(self) -> None:
        mock_config: MagicMock = MagicMock()
        mock_config.app.domain = "example.com"
        mock_config.app.title = "Example App"
        with patch("acb.depends.depends.__call__", return_value=mock_config):
            settings: SmtpBaseSettings = SmtpBaseSettings(
                port=587,
                tls=True,
                ssl=False,
                password=SecretStr("password"),
                forwards={"admin": "pat@example.com"},
                domain="mail.example.com",
                default_from="info@example.com",
                default_from_name="Example App",
            )
            assert settings.domain == "mail.example.com"
            assert settings.default_from == "info@example.com"
            assert settings.default_from_name == "Example App"
            assert settings.port == 587
            assert settings.tls
            assert not settings.ssl
            assert isinstance(settings.password, SecretStr)
            assert "admin" in settings.forwards
            assert settings.forwards["admin"] == "pat@example.com"

    def test_email_validation(self) -> None:
        settings = SmtpBaseSettings(
            domain="example.com",
            default_from="valid@example.com",
        )
        assert settings.default_from == "valid@example.com"


class TestSmtpBase:
    @pytest.fixture
    def smtp_base(self) -> MockSmtpBase:
        smtp_base: MockSmtpBase = MockSmtpBase()
        return smtp_base

    @pytest.mark.asyncio
    async def test_init(self, smtp_base: MockSmtpBase) -> None:
        assert not smtp_base.initialized
        await smtp_base.init()
        assert smtp_base.initialized



"""Tests for the ACB vector base module."""

from unittest.mock import AsyncMock, Mock, patch

import pytest
from pydantic import SecretStr

from acb.adapters.vector._base import (
    VectorBase,
    VectorBaseSettings,
    VectorCollection,
    VectorDocument,
    VectorSearchResult,
)
from acb.config import AppSettings, Config, DebugSettings


class TestVectorSearchResult:
    """Test the VectorSearchResult model."""

    def test_vector_search_result_defaults(self) -> None:
        """Test VectorSearchResult default values."""
        result = VectorSearchResult(id="test-id", score=0.95)

        assert result.id == "test-id"
        assert result.score == 0.95
        assert result.metadata == {}
        assert result.vector is None

    def test_vector_search_result_with_values(self) -> None:
        """Test VectorSearchResult with custom values."""
        result = VectorSearchResult(
            id="test-id", score=0.85, metadata={"key": "value"}, vector=[1.0, 2.0, 3.0]
        )

        assert result.id == "test-id"
        assert result.score == 0.85
        assert result.metadata == {"key": "value"}
        assert result.vector == [1.0, 2.0, 3.0]


class TestVectorDocument:
    """Test the VectorDocument model."""

    def test_vector_document_defaults(self) -> None:
        """Test VectorDocument default values."""
        doc = VectorDocument(id="test-id", vector=[1.0, 2.0, 3.0])

        assert doc.id == "test-id"
        assert doc.vector == [1.0, 2.0, 3.0]
        assert doc.metadata == {}

    def test_vector_document_with_none_id(self) -> None:
        """Test VectorDocument with None id."""
        doc = VectorDocument(vector=[1.0, 2.0, 3.0])

        assert doc.id is None
        assert doc.vector == [1.0, 2.0, 3.0]
        assert doc.metadata == {}

    def test_vector_document_with_metadata(self) -> None:
        """Test VectorDocument with metadata."""
        doc = VectorDocument(
            id="test-id",
            vector=[1.0, 2.0, 3.0],
            metadata={"key": "value", "category": "test"},
        )

        assert doc.id == "test-id"
        assert doc.vector == [1.0, 2.0, 3.0]
        assert doc.metadata == {"key": "value", "category": "test"}


class TestVectorBaseSettings:
    """Test the VectorBaseSettings model."""

    def test_vector_base_settings_defaults(self) -> None:
        """Test VectorBaseSettings default values."""
        settings = VectorBaseSettings()

        assert settings.host.get_secret_value() == "127.0.0.1"
        assert settings.port is None
        assert settings.collection_prefix == ""
        assert settings.default_dimension == 1536
        assert settings.default_distance_metric == "cosine"
        assert settings.connect_timeout == 30.0
        assert settings.request_timeout == 30.0
        assert settings.max_retries == 3
        assert settings.batch_size == 100
        assert settings.max_connections == 10
        assert settings.enable_caching is True
        assert settings.enable_hybrid_search is False
        assert settings.enable_auto_scaling is False
        assert settings.enable_connection_pooling is True

    def test_vector_base_settings_custom_values(self) -> None:
        """Test VectorBaseSettings with custom values."""
        settings = VectorBaseSettings(
            host=SecretStr("vector-db.example.com"),
            port=6379,
            collection_prefix="prod_",
            default_dimension=1024,
            default_distance_metric="euclidean",
            connect_timeout=60.0,
            request_timeout=60.0,
            max_retries=5,
            batch_size=200,
            max_connections=20,
            enable_caching=False,
            enable_hybrid_search=True,
            enable_auto_scaling=True,
            enable_connection_pooling=False,
        )

        assert settings.host.get_secret_value() == "vector-db.example.com"
        assert settings.port == 6379
        assert settings.collection_prefix == "prod_"
        assert settings.default_dimension == 1024
        assert settings.default_distance_metric == "euclidean"
        assert settings.connect_timeout == 60.0
        assert settings.request_timeout == 60.0
        assert settings.max_retries == 5
        assert settings.batch_size == 200
        assert settings.max_connections == 20
        assert settings.enable_caching is False
        assert settings.enable_hybrid_search is True
        assert settings.enable_auto_scaling is True
        assert settings.enable_connection_pooling is False


class TestVectorCollection:
    """Test the VectorCollection class."""

    def test_vector_collection_initialization(self) -> None:
        """Test VectorCollection initialization."""
        mock_adapter = Mock()
        collection = VectorCollection(mock_adapter, "test_collection")

        assert collection.adapter == mock_adapter
        assert collection.name == "test_collection"

    @pytest.mark.asyncio
    async def test_vector_collection_search(self) -> None:
        """Test VectorCollection search method."""
        mock_adapter = Mock()
        mock_adapter.search = AsyncMock(return_value=[Mock()])
        collection = VectorCollection(mock_adapter, "test_collection")

        result = await collection.search([1.0, 2.0, 3.0], limit=5)

        mock_adapter.search.assert_called_once_with(
            "test_collection", [1.0, 2.0, 3.0], 5, None, False
        )
        assert result == [Mock()]

    @pytest.mark.asyncio
    async def test_vector_collection_insert(self) -> None:
        """Test VectorCollection insert method."""
        mock_adapter = Mock()
        mock_adapter.insert = AsyncMock(return_value=["id1", "id2"])
        collection = VectorCollection(mock_adapter, "test_collection")

        mock_documents = [Mock(), Mock()]
        result = await collection.insert(mock_documents)

        mock_adapter.insert.assert_called_once_with("test_collection", mock_documents)
        assert result == ["id1", "id2"]

    @pytest.mark.asyncio
    async def test_vector_collection_upsert(self) -> None:
        """Test VectorCollection upsert method."""
        mock_adapter = Mock()
        mock_adapter.upsert = AsyncMock(return_value=["id1", "id2"])
        collection = VectorCollection(mock_adapter, "test_collection")

        mock_documents = [Mock(), Mock()]
        result = await collection.upsert(mock_documents)

        mock_adapter.upsert.assert_called_once_with("test_collection", mock_documents)
        assert result == ["id1", "id2"]

    @pytest.mark.asyncio
    async def test_vector_collection_delete(self) -> None:
        """Test VectorCollection delete method."""
        mock_adapter = Mock()
        mock_adapter.delete = AsyncMock(return_value=True)
        collection = VectorCollection(mock_adapter, "test_collection")

        result = await collection.delete(["id1", "id2"])

        mock_adapter.delete.assert_called_once_with("test_collection", ["id1", "id2"])
        assert result is True

    @pytest.mark.asyncio
    async def test_vector_collection_get(self) -> None:
        """Test VectorCollection get method."""
        mock_adapter = Mock()
        mock_adapter.get = AsyncMock(return_value=[Mock()])
        collection = VectorCollection(mock_adapter, "test_collection")

        result = await collection.get(["id1", "id2"])

        mock_adapter.get.assert_called_once_with(
            "test_collection", ["id1", "id2"], False
        )
        assert result == [Mock()]

    @pytest.mark.asyncio
    async def test_vector_collection_count(self) -> None:
        """Test VectorCollection count method."""
        mock_adapter = Mock()
        mock_adapter.count = AsyncMock(return_value=42)
        collection = VectorCollection(mock_adapter, "test_collection")

        result = await collection.count()

        mock_adapter.count.assert_called_once_with("test_collection", None)
        assert result == 42


class TestVectorBase:
    """Test the VectorBase class."""

    @pytest.fixture
    def mock_config(self) -> Mock:
        """Create a mock config."""
        mock_config = Mock(spec=Config)
        mock_config.app = Mock(spec=AppSettings)
        mock_config.app.name = "test_app"
        mock_config.debug = Mock(spec=DebugSettings)
        mock_config.debug.sql = False
        return mock_config

    @pytest.fixture
    def vector_base(self, mock_config: Mock) -> VectorBase:
        """Create a VectorBase instance with mock config."""
        with patch("acb.adapters.vector._base.depends.get", return_value=mock_config):
            vector = VectorBase()
            vector.config = mock_config
            vector.settings = VectorBaseSettings()
            vector.logger = Mock()
            return vector

    def test_vector_base_initialization(self, vector_base: VectorBase) -> None:
        """Test VectorBase initialization."""
        assert vector_base._collections == {}
        assert vector_base._client is None
        assert vector_base._cache is None
        assert vector_base._hybrid_search is None
        assert vector_base._auto_scaler is None
        assert vector_base._connection_pool is None

    def test_vector_base_dynamic_collection_access(
        self, vector_base: VectorBase
    ) -> None:
        """Test VectorBase dynamic collection access."""
        # First access should create a new collection
        collection1 = vector_base.test_collection
        assert isinstance(collection1, VectorCollection)
        assert collection1.name == "test_collection"
        assert collection1.adapter == vector_base

        # Second access should return the same collection
        collection2 = vector_base.test_collection
        assert collection1 is collection2

    @pytest.mark.asyncio
    async def test_vector_base_get_client(self, vector_base: VectorBase) -> None:
        """Test VectorBase get_client method."""
        mock_client = Mock()
        with patch.object(
            vector_base, "_ensure_client", AsyncMock(return_value=mock_client)
        ):
            client = await vector_base.get_client()
            assert client == mock_client

    @pytest.mark.asyncio
    async def test_vector_base_get_cache_disabled(
        self, vector_base: VectorBase
    ) -> None:
        """Test VectorBase get_cache when caching is disabled."""
        vector_base.settings.enable_caching = False

        cache = await vector_base.get_cache()
        assert cache is None

    @pytest.mark.asyncio
    async def test_vector_base_get_cache_enabled(self, vector_base: VectorBase) -> None:
        """Test VectorBase get_cache when caching is enabled."""
        vector_base.settings.enable_caching = True
        vector_base._cache = Mock()

        cache = await vector_base.get_cache()
        assert cache == vector_base._cache

    @pytest.mark.asyncio
    async def test_vector_base_get_hybrid_search_disabled(
        self, vector_base: VectorBase
    ) -> None:
        """Test VectorBase get_hybrid_search when disabled."""
        vector_base.settings.enable_hybrid_search = False

        hybrid_search = await vector_base.get_hybrid_search()
        assert hybrid_search is None

    @pytest.mark.asyncio
    async def test_vector_base_get_hybrid_search_enabled(
        self, vector_base: VectorBase
    ) -> None:
        """Test VectorBase get_hybrid_search when enabled."""
        vector_base.settings.enable_hybrid_search = True
        vector_base._hybrid_search = Mock()

        hybrid_search = await vector_base.get_hybrid_search()
        assert hybrid_search == vector_base._hybrid_search

    @pytest.mark.asyncio
    async def test_vector_base_get_auto_scaler_disabled(
        self, vector_base: VectorBase
    ) -> None:
        """Test VectorBase get_auto_scaler when disabled."""
        vector_base.settings.enable_auto_scaling = False

        auto_scaler = await vector_base.get_auto_scaler()
        assert auto_scaler is None

    @pytest.mark.asyncio
    async def test_vector_base_get_auto_scaler_enabled(
        self, vector_base: VectorBase
    ) -> None:
        """Test VectorBase get_auto_scaler when enabled."""
        vector_base.settings.enable_auto_scaling = True
        vector_base._auto_scaler = Mock()

        auto_scaler = await vector_base.get_auto_scaler()
        assert auto_scaler == vector_base._auto_scaler

    @pytest.mark.asyncio
    async def test_vector_base_get_connection_pool_disabled(
        self, vector_base: VectorBase
    ) -> None:
        """Test VectorBase get_connection_pool when disabled."""
        vector_base.settings.enable_connection_pooling = False

        connection_pool = await vector_base.get_connection_pool()
        assert connection_pool is None

    @pytest.mark.asyncio
    async def test_vector_base_get_connection_pool_enabled(
        self, vector_base: VectorBase
    ) -> None:
        """Test VectorBase get_connection_pool when enabled."""
        vector_base.settings.enable_connection_pooling = True
        vector_base._connection_pool = Mock()

        connection_pool = await vector_base.get_connection_pool()
        assert connection_pool == vector_base._connection_pool

    @pytest.mark.asyncio
    async def test_vector_base_search_with_cache(self, vector_base: VectorBase) -> None:
        """Test VectorBase search_with_cache method."""
        # Test with cache disabled - should fallback to regular search
        vector_base.settings.enable_caching = False
        with patch.object(vector_base, "search", AsyncMock(return_value=[Mock()])):
            result = await vector_base.search_with_cache(
                "test_collection", [1.0, 2.0, 3.0], limit=5
            )
            assert len(result) == 1
            vector_base.search.assert_called_once()

    @pytest.mark.asyncio
    async def test_vector_base_hybrid_search(self, vector_base: VectorBase) -> None:
        """Test VectorBase hybrid_search method."""
        # Test with hybrid search disabled - should fallback to regular search
        vector_base.settings.enable_hybrid_search = False
        with patch.object(vector_base, "search", AsyncMock(return_value=[Mock()])):
            result = await vector_base.hybrid_search(
                "test_collection", [1.0, 2.0, 3.0], "test query", limit=5
            )
            assert len(result) == 1
            vector_base.search.assert_called_once()

    @pytest.mark.asyncio
    async def test_vector_base_transaction_context_manager(
        self, vector_base: VectorBase
    ) -> None:
        """Test VectorBase transaction context manager."""
        mock_client = Mock()
        with patch.object(
            vector_base, "get_client", AsyncMock(return_value=mock_client)
        ):
            async with vector_base.transaction() as client:
                assert client == mock_client

    def test_vector_base_abstract_methods(self) -> None:
        """Test that VectorBase has abstract methods."""
        # VectorBase should not be instantiable directly because of abstract methods
        # But we can test that the methods exist
        assert hasattr(VectorBase, "init")
        # The method should be abstract
        import inspect

        assert inspect.iscoroutinefunction(VectorBase.init)


# Additional edge case tests
class TestVectorBaseEdgeCases:
    """Test edge cases for vector base classes."""

    def test_vector_document_id_none_explicit(self) -> None:
        """Test VectorDocument with explicit None id."""
        doc = VectorDocument(id=None, vector=[1.0, 2.0, 3.0])
        assert doc.id is None

    def test_vector_search_result_empty_metadata(self) -> None:
        """Test VectorSearchResult with explicitly empty metadata."""
        result = VectorSearchResult(id="test", score=0.5, metadata={})
        assert result.metadata == {}
        assert result.vector is None

    def test_vector_search_result_with_vector(self) -> None:
        """Test VectorSearchResult with vector data."""
        result = VectorSearchResult(
            id="test", score=0.75, metadata={"category": "test"}, vector=[0.1, 0.2, 0.3]
        )
        assert result.vector == [0.1, 0.2, 0.3]

    @pytest.mark.asyncio
    async def test_vector_collection_search_with_filters(self) -> None:
        """Test VectorCollection search with filters."""
        mock_adapter = Mock()
        mock_adapter.search = AsyncMock(return_value=[])
        collection = VectorCollection(mock_adapter, "test_collection")

        filters = {"category": "test", "status": "active"}
        await collection.search([1.0, 2.0], filter_expr=filters, include_vectors=True)

        mock_adapter.search.assert_called_once_with(
            "test_collection", [1.0, 2.0], 10, filters, True
        )

    @pytest.mark.asyncio
    async def test_vector_collection_get_with_vectors(self) -> None:
        """Test VectorCollection get with include_vectors=True."""
        mock_adapter = Mock()
        mock_adapter.get = AsyncMock(return_value=[])
        collection = VectorCollection(mock_adapter, "test_collection")

        await collection.get(["id1"], include_vectors=True)

        mock_adapter.get.assert_called_once_with("test_collection", ["id1"], True)

    def test_vector_base_settings_ssl_config_mixin(self) -> None:
        """Test that VectorBaseSettings inherits from SSLConfigMixin."""
        from acb.ssl_config import SSLConfigMixin

        assert issubclass(VectorBaseSettings, SSLConfigMixin)

    def test_vector_base_settings_secrets(self) -> None:
        """Test that VectorBaseSettings handles secrets correctly."""
        settings = VectorBaseSettings()
        # Host should be a SecretStr
        assert isinstance(settings.host, SecretStr)
        assert settings.host.get_secret_value() == "127.0.0.1"



from __future__ import annotations

import time
from enum import Enum

import typing as t
from dataclasses import dataclass, field
from datetime import UTC, datetime


class HealthStatus(str, Enum):
    HEALTHY = "healthy"
    DEGRADED = "degraded"
    UNHEALTHY = "unhealthy"
    CRITICAL = "critical"
    UNKNOWN = "unknown"

    def __str__(self) -> str:  # pragma: no cover - trivial
        return str(self.value)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, HealthStatus):
            return NotImplemented
        order = {
            HealthStatus.HEALTHY: 0,
            HealthStatus.DEGRADED: 1,
            HealthStatus.UNHEALTHY: 2,
            HealthStatus.CRITICAL: 3,
            HealthStatus.UNKNOWN: 4,
        }
        return order[self] < order[other]

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, HealthStatus):
            return NotImplemented
        order = {
            HealthStatus.HEALTHY: 0,
            HealthStatus.DEGRADED: 1,
            HealthStatus.UNHEALTHY: 2,
            HealthStatus.CRITICAL: 3,
            HealthStatus.UNKNOWN: 4,
        }
        return order[self] > order[other]


@dataclass
class ComponentHealth:
    name: str
    status: HealthStatus
    message: str | None = None
    latency_ms: float | None = None
    metadata: dict[str, t.Any] = field(default_factory=dict)
    check_type: str | None = None
    component_id: str | None = None
    component_name: str | None = None
    details: dict[str, t.Any] = field(default_factory=dict)
    timestamp: float | None = None
    duration_ms: float | None = None
    error: str | None = None

    def to_dict(self) -> dict[str, t.Any]:
        result: dict[str, t.Any] = {"name": self.name, "status": self.status.value}
        if self.message is not None:
            result["message"] = self.message
        if self.latency_ms is not None:
            result["latency_ms"] = round(self.latency_ms, 2)
        if self.metadata:
            result["metadata"] = self.metadata
        if self.check_type is not None:
            result["check_type"] = self.check_type
        if self.component_id is not None:
            result["component_id"] = self.component_id
        if self.component_name is not None:
            result["component_name"] = self.component_name
        if self.details:
            result["details"] = self.details
        if self.timestamp is not None:
            result["timestamp"] = self.timestamp
        if self.duration_ms is not None:
            result["duration_ms"] = round(self.duration_ms, 2)
        if self.error is not None:
            result["error"] = self.error
        return result


@dataclass
class HealthCheckResponse:
    status: HealthStatus
    timestamp: str
    version: str
    components: list[ComponentHealth]
    uptime_seconds: float
    metadata: dict[str, t.Any] = field(default_factory=dict)

    @classmethod
    def create(
        cls,
        components: list[ComponentHealth],
        *,
        version: str,
        start_time: float,
        metadata: dict[str, t.Any] | None = None,
    ) -> HealthCheckResponse:
        overall = (
            HealthStatus.HEALTHY
            if not components
            else max(c.status for c in components)
        )
        return cls(
            status=overall,
            timestamp=datetime.now(UTC).isoformat(),
            version=version,
            components=components,
            uptime_seconds=time.time() - start_time,
            metadata=metadata or {},
        )

    def to_dict(self) -> dict[str, t.Any]:
        result: dict[str, t.Any] = {
            "status": self.status.value,
            "timestamp": self.timestamp,
            "version": self.version,
            "uptime_seconds": round(self.uptime_seconds, 2),
            "components": [c.to_dict() for c in self.components],
        }
        if self.metadata:
            result["metadata"] = self.metadata
        return result

    def is_healthy(self) -> bool:  # pragma: no cover - trivial
        return self.status == HealthStatus.HEALTHY

    def is_ready(self) -> bool:  # pragma: no cover - trivial
        return self.status != HealthStatus.UNHEALTHY


__all__ = [
    "HealthStatus",
    "ComponentHealth",
    "HealthCheckResponse",
]



"""Tests for basic template rendering functionality."""

from __future__ import annotations

import pytest
from datetime import datetime


class TestBasicRendering:
    """Test basic template rendering operations."""

    @pytest.mark.asyncio
    async def test_render_simple_template(self, templates, sample_template):
        """Test rendering a simple template with variable substitution."""
        result = await templates.render("sample.html", name="World")
        assert result == "Hello World!"

    @pytest.mark.asyncio
    async def test_render_with_context_dict(self, templates, sample_template):
        """Test rendering with context dictionary."""
        context = {"name": "Alice"}
        result = await templates.render("sample.html", context=context)
        assert result == "Hello Alice!"

    @pytest.mark.asyncio
    async def test_render_with_kwargs(self, templates, sample_template):
        """Test rendering with kwargs."""
        result = await templates.render("sample.html", name="Bob")
        assert result == "Hello Bob!"

    @pytest.mark.asyncio
    async def test_render_merges_context_and_kwargs(self, templates, sample_template):
        """Test that context and kwargs are merged."""
        context = {"name": "Charlie"}
        # kwargs should override context
        result = await templates.render("sample.html", context=context, name="David")
        assert result == "Hello David!"

    @pytest.mark.asyncio
    async def test_render_string_basic(self, templates):
        """Test rendering a template string."""
        result = await templates.render_string("Hello {{ name }}!", name="World")
        assert result == "Hello World!"

    @pytest.mark.asyncio
    async def test_render_string_with_context(self, templates):
        """Test rendering string with context dictionary."""
        context = {"greeting": "Hi", "name": "Alice"}
        result = await templates.render_string(
            "{{ greeting }} {{ name }}!", context=context
        )
        assert result == "Hi Alice!"

    @pytest.mark.asyncio
    async def test_template_not_found(self, templates):
        """Test error handling when template doesn't exist."""
        from jinja2 import TemplateNotFound

        with pytest.raises(TemplateNotFound):
            await templates.render("nonexistent.html")

    @pytest.mark.asyncio
    async def test_template_with_no_variables(self, templates, template_dir):
        """Test rendering template with no variables."""
        static_template = template_dir / "static.html"
        static_template.write_text("<h1>Static Content</h1>")

        result = await templates.render("static.html")
        assert result == "<h1>Static Content</h1>"

    @pytest.mark.asyncio
    async def test_template_with_missing_variable(self, templates, sample_template):
        """Test rendering with missing variable (Jinja2 default: renders as empty string)."""
        result = await templates.render("sample.html")
        # Jinja2's default Undefined renders as empty string ''
        assert result == "Hello !"


class TestTemplateInheritance:
    """Test template inheritance and blocks."""

    @pytest.mark.asyncio
    async def test_basic_inheritance(self, templates, complex_template):
        """Test basic template inheritance."""
        from datetime import datetime

        result = await templates.render(
            "child.html",
            title="Test Page",
            description="This is a test",
            created_at=datetime(2025, 1, 25, 14, 30, 0),
            file_size=1536,
            data={"key": "value"},
        )

        assert "<title>Test Page</title>" in result
        assert "<h1>Test Page</h1>" in result
        assert "<p>This is a test</p>" in result
        assert "2025-01-25 14:30:00" in result
        assert "1.5 KiB" in result
        assert '"key": "value"' in result

    @pytest.mark.asyncio
    async def test_email_template_inheritance(self, templates, email_templates):
        """Test email template structure with inheritance."""
        result = await templates.render(
            "email/welcome.html",
            site_name="Test Site",
            year=2025,
            user={"name": "Alice", "email": "alice@example.com"},
            joined_at=datetime(2025, 1, 25),
        )

        assert "Test Site" in result
        assert "Welcome, Alice!" in result
        assert "alice@example.com" in result
        assert "January 25, 2025" in result
        assert "&copy; 2025 Test Site" in result


class TestAutoEscaping:
    """Test HTML auto-escaping behavior."""

    @pytest.mark.asyncio
    async def test_auto_escaping_enabled_by_default(self, templates):
        """Test that auto-escaping is enabled by default."""
        html = "<script>alert('XSS')</script>"
        result = await templates.render_string("{{ content }}", content=html)
        assert "&lt;script&gt;" in result
        assert "&lt;/script&gt;" in result
        assert "<script>" not in result

    @pytest.mark.asyncio
    async def test_safe_filter_disables_escaping(self, templates):
        """Test that |safe filter disables escaping."""
        html = "<strong>Bold</strong>"
        result = await templates.render_string("{{ content|safe }}", content=html)
        assert "<strong>Bold</strong>" in result

    @pytest.mark.asyncio
    async def test_auto_escaping_disabled(self, template_dir):
        """Test templates with auto-escaping disabled."""
        from acb.adapters.templates import TemplatesAdapter

        templates_no_escape = TemplatesAdapter(
            template_dir=template_dir, autoescape=False
        )

        html = "<script>alert('test')</script>"
        result = await templates_no_escape.render_string("{{ content }}", content=html)
        assert "<script>alert('test')</script>" in result
        assert "&lt;script&gt;" not in result


class TestTemplateSettings:
    """Test template adapter initialization and settings."""

    def test_default_settings(self, tmp_path):
        """Test adapter with default settings."""
        from acb.adapters.templates import TemplatesAdapter

        templates = TemplatesAdapter(template_dir=tmp_path / "templates")

        assert templates.settings.enable_async is True
        assert templates.settings.autoescape is True
        assert templates.settings.cache_size == 400
        assert templates.settings.auto_reload is True
        assert templates.template_dir == tmp_path / "templates"

    def test_custom_settings(self, tmp_path):
        """Test adapter with custom settings."""
        from acb.adapters.templates import TemplatesAdapter

        templates = TemplatesAdapter(
            template_dir=tmp_path / "custom",
            enable_async=False,
            autoescape=False,
            cache_size=200,
            auto_reload=False,
        )

        assert templates.settings.enable_async is False
        assert templates.settings.autoescape is False
        assert templates.settings.cache_size == 200
        assert templates.settings.auto_reload is False

    def test_settings_object(self, tmp_path):
        """Test adapter with settings object."""
        from acb.adapters.templates import TemplatesAdapter, TemplatesBaseSettings

        settings = TemplatesBaseSettings(
            template_dir=tmp_path / "templates",
            cache_size=500,
        )

        templates = TemplatesAdapter(settings=settings)
        assert templates.settings.cache_size == 500

    def test_template_directory_created(self, tmp_path):
        """Test that template directory is created if it doesn't exist."""
        from acb.adapters.templates import TemplatesAdapter

        new_dir = tmp_path / "new_templates"
        assert not new_dir.exists()

        TemplatesAdapter(template_dir=new_dir)
        assert new_dir.exists()
        assert new_dir.is_dir()



"""DuckDB PGQ graph adapter."""

from __future__ import annotations

import json
import re
from pathlib import Path
from uuid import UUID, uuid4

import typing as t
from contextlib import asynccontextmanager, suppress
from dataclasses import dataclass
from datetime import UTC, datetime
from pydantic import Field
from sqlalchemy import text
from sqlalchemy.engine import URL
from sqlalchemy.engine.url import make_url
from sqlalchemy.ext.asyncio import (
    AsyncConnection,
    AsyncEngine,
    AsyncTransaction,
    create_async_engine,
)

from acb.adapters import AdapterCapability, AdapterMetadata, AdapterStatus
from acb.adapters.graph._base import (
    GraphBase,
    GraphBaseSettings,
    GraphEdgeModel,
    GraphNodeModel,
    GraphPathModel,
    GraphQueryLanguage,
    GraphQueryResult,
    GraphSchemaModel,
    GraphTraversalDirection,
)
from acb.depends import depends

MODULE_METADATA = AdapterMetadata(
    module_id=UUID("019a1c00-e6f0-7d4a-a749-b305c0b2a4d2"),
    name="DuckDB PGQ",
    category="graph",
    provider="duckdb",
    version="0.1.0",
    acb_min_version="0.19.0",
    author="lesleslie <les@wedgwoodwebworks.com>",
    created_date="2025-02-15",
    last_modified="2025-02-15",
    status=AdapterStatus.BETA,
    capabilities=[
        AdapterCapability.ASYNC_OPERATIONS,
        AdapterCapability.TRANSACTIONS,
        AdapterCapability.TLS_SUPPORT,
        AdapterCapability.GRAPH_TRAVERSAL,
        AdapterCapability.BULK_OPERATIONS,
        AdapterCapability.GRAPH_SCHEMA_VALIDATION,
        AdapterCapability.PATHFINDING,
    ],
    required_packages=[
        "duckdb>=0.10.2",
        "duckdb-engine>=0.11.2",
        "sqlalchemy>=2.0.41",
    ],
    optional_packages={
        "duckdb-engine[arrow]": "Enable Arrow acceleration for analytical queries",
    },
    description=(
        "DuckDB graph adapter powered by the DuckPGQ extension. "
        "Suitable for lightweight property graph workloads and embedded analytics."
    ),
    settings_class="DuckDBPGQSettings",
)


def _safe_ident(name: str) -> str:
    """Validate and return a safe SQL identifier.

    Allows only letters, numbers, and underscores, starting with a letter or underscore.
    Raises ValueError if the identifier is unsafe.
    """
    if not re.match(
        r"^[A-Za-z_][A-Za-z0-9_]*$", name
    ):  # REGEX OK: SQL identifier validation - anchored character classes only  # nosec B108
        msg = f"Unsafe SQL identifier: {name!r}"
        raise ValueError(msg)
    return name


def _ensure_directory(target: Path) -> None:
    if target.name == ":memory:":
        return
    if not target.parent.exists():
        target.parent.mkdir(parents=True, exist_ok=True)


class DuckDBPGQSettings(GraphBaseSettings):
    """DuckDB PGQ specific settings."""

    database_url: str = "duckdb:///data/graph.duckdb"
    graph_name: str = "acb_graph"
    nodes_table: str = "acb_graph_nodes"
    edges_table: str = "acb_graph_edges"
    install_extensions: list[str] = Field(default_factory=lambda: ["duckpgq"])
    pragmas: dict[str, str] = Field(default_factory=dict)

    _url: URL | None = None

    def __init__(self, **values: t.Any) -> None:  # type: ignore[override]
        super().__init__(**values)
        self._prepare_url()

    def _prepare_url(self) -> None:
        url = make_url(self.database_url)
        if url.database and url.database != ":memory:":
            _ensure_directory(Path(url.database))
        self._url = url

    @property
    def database_path(self) -> Path | None:
        if not self._url or not self._url.database or self._url.database == ":memory:":
            return None
        return Path(self._url.database)


@dataclass(slots=True)
class _DuckDBTransaction:
    connection: AsyncConnection
    transaction: AsyncTransaction


class Graph(GraphBase):
    """DuckDB PGQ graph adapter."""

    def __init__(
        self,
        settings: DuckDBPGQSettings | None = None,
        **kwargs: t.Any,
    ) -> None:
        super().__init__(**kwargs)
        if settings is not None:
            self._settings = settings
        else:
            self._settings = DuckDBPGQSettings(**kwargs)
        self._engine: AsyncEngine | None = None

    @property
    def query_language(self) -> GraphQueryLanguage:
        return GraphQueryLanguage.PGQ

    @property
    def supported_features(self) -> list[str]:
        return [
            "pgq_queries",
            "transactions",
            "bulk_ingest",
            "json_properties",
            "in_memory_mode",
            "analytics",
        ]

    async def _create_client(self) -> AsyncEngine:
        assert self._settings._url is not None
        engine = create_async_engine(
            str(self._settings._url),
            pool_pre_ping=False,
            future=True,
        )
        async with engine.begin() as conn:
            await self._install_extensions(conn)
            await self._configure_pragmas(conn)
            await self._ensure_tables(conn)
            await self._register_graph(conn)
        self._engine = engine
        return engine

    async def _install_extensions(self, conn: AsyncConnection) -> None:
        for extension in self._settings.install_extensions:
            safe_extension = _safe_ident(extension)
            stmt = text(f"INSTALL {safe_extension}")  # nosec B608
            with suppress(Exception):
                await conn.execute(stmt)
            stmt = text(f"LOAD {safe_extension}")  # nosec B608
            with suppress(Exception):
                await conn.execute(stmt)

    async def _configure_pragmas(self, conn: AsyncConnection) -> None:
        for pragma, value in self._settings.pragmas.items():
            value_str = value if isinstance(value, str) else format(value)
            numeric_check = value_str.lstrip("-+").replace(".", "", 1).isdigit()
            if numeric_check:
                clause = value_str
            else:
                escaped = value_str.replace("'", "''")
                clause = f"'{escaped}'"
            safe_pragma = _safe_ident(pragma)
            stmt = text(f"PRAGMA {safe_pragma}={clause}")  # nosec B608
            await conn.execute(stmt)

    async def _ensure_tables(self, conn: AsyncConnection) -> None:
        nodes_table = _safe_ident(self._settings.nodes_table)
        edges_table = _safe_ident(self._settings.edges_table)
        sql_nodes = f"""
                CREATE TABLE IF NOT EXISTS {nodes_table} (
                    id TEXT PRIMARY KEY,
                    labels JSON,
                    properties JSON,
                    created_at TIMESTAMP,
                    updated_at TIMESTAMP
                )
                """  # nosec B608
        await conn.execute(text(sql_nodes))
        await conn.execute(
            text(
                f"""
                CREATE TABLE IF NOT EXISTS {edges_table} (
                    id TEXT PRIMARY KEY,
                    type TEXT,
                    from_node TEXT,
                    to_node TEXT,
                    properties JSON,
                    created_at TIMESTAMP,
                    updated_at TIMESTAMP
                )
                """
                # nosec B608
            )
        )
        await conn.execute(
            text(
                f"CREATE INDEX IF NOT EXISTS idx_{edges_table}_from ON {edges_table}(from_node)"
                # nosec B608
            )
        )
        await conn.execute(
            text(
                f"CREATE INDEX IF NOT EXISTS idx_{edges_table}_to ON {edges_table}(to_node)"
                # nosec B608
            )
        )

    async def _register_graph(self, conn: AsyncConnection) -> None:
        graph_name = self._settings.graph_name
        nodes_table = self._settings.nodes_table
        edges_table = self._settings.edges_table
        create_stmt = text(
            """
            CALL pgq.create_graph(
                graph_name => :graph_name,
                nodes_table => :nodes_table,
                edges_table => :edges_table,
                node_id => 'id',
                from_node => 'from_node',
                to_node => 'to_node'
            )
            """
        )
        with suppress(Exception):
            await conn.execute(
                create_stmt,
                {
                    "graph_name": graph_name,
                    "nodes_table": nodes_table,
                    "edges_table": edges_table,
                },
            )

    @asynccontextmanager
    async def _connection(self) -> t.AsyncIterator[AsyncConnection]:
        if isinstance(self._transaction, _DuckDBTransaction):
            yield self._transaction.connection
            return
        engine = await self._ensure_client()
        async with engine.begin() as conn:
            yield conn

    async def _begin_transaction(self, client: AsyncEngine) -> _DuckDBTransaction:
        conn = await client.connect()
        transaction = await conn.begin()
        return _DuckDBTransaction(connection=conn, transaction=transaction)

    async def _commit_transaction(self, transaction: _DuckDBTransaction) -> None:
        await transaction.transaction.commit()
        await transaction.connection.close()

    async def _rollback_transaction(self) -> None:
        if isinstance(self._transaction, _DuckDBTransaction):
            await self._transaction.transaction.rollback()
            await self._transaction.connection.close()

    async def _execute_query(
        self,
        query: str,
        parameters: dict[str, t.Any] | None = None,
        timeout: float | None = None,
    ) -> GraphQueryResult:
        params = parameters or {}
        async with self._connection() as conn:
            if timeout:
                await conn.execute(
                    text(f"PRAGMA statement_timeout={int(timeout * 1000)}")
                )
            stripped = query.lstrip()
            if stripped.upper().startswith("MATCH"):
                stmt = text("SELECT * FROM GRAPH_QUERY(:graph_name, :pgq_query)")
                result = await conn.execute(
                    stmt,
                    {
                        "graph_name": self._settings.graph_name,
                        "pgq_query": query,
                    },
                )
            else:
                stmt = text(query)
                result = await conn.execute(stmt, params)
            rows = result.mappings().all()
        records = [dict(row) for row in rows]
        return GraphQueryResult(
            nodes=[],
            edges=[],
            paths=[],
            records=records,
            metadata={
                "graph": self._settings.graph_name,
                "rowcount": len(records),
            },
            query_language=GraphQueryLanguage.PGQ
            if stripped.upper().startswith("MATCH")
            else None,
        )

    async def _create_node(
        self,
        labels: list[str],
        properties: dict[str, t.Any],
    ) -> GraphNodeModel:
        node_id = str(uuid4())
        now = datetime.now(tz=UTC)
        async with self._connection() as conn:
            sql = (
                "INSERT INTO "
                + _safe_ident(self._settings.nodes_table)
                + " (id, labels, properties, created_at, updated_at) "
                + "VALUES (:id, :labels, :properties, :created_at, :updated_at)"
            )  # nosec B608
            await conn.execute(
                text(sql),
                {
                    "id": node_id,
                    "labels": json.dumps(labels),
                    "properties": json.dumps(properties),
                    "created_at": now,
                    "updated_at": now,
                },
            )
        return GraphNodeModel(
            id=node_id,
            labels=labels,
            properties=properties,
            created_at=now,
            updated_at=now,
        )

    async def _get_node(self, node_id: str) -> GraphNodeModel | None:
        async with self._connection() as conn:
            sql = (
                "SELECT id, labels, properties, created_at, updated_at FROM "
                + _safe_ident(self._settings.nodes_table)
                + " WHERE id = :id"
            )  # nosec B608
            result = await conn.execute(
                text(sql),
                {"id": node_id},
            )
            row = result.mappings().first()
        if not row:
            return None
        return GraphNodeModel(
            id=row["id"],
            labels=json.loads(row["labels"]) if row["labels"] else [],
            properties=json.loads(row["properties"]) if row["properties"] else {},
            created_at=row["created_at"],
            updated_at=row["updated_at"],
        )

    async def _update_node(
        self,
        node_id: str,
        properties: dict[str, t.Any],
    ) -> GraphNodeModel:
        now = datetime.now(tz=UTC)
        async with self._connection() as conn:
            sql = (
                "UPDATE "
                + _safe_ident(self._settings.nodes_table)
                + " SET properties = :properties, updated_at = :updated_at WHERE id = :id"
            )  # nosec B608
            await conn.execute(
                text(sql),
                {
                    "id": node_id,
                    "properties": json.dumps(properties),
                    "updated_at": now,
                },
            )
        node = await self._get_node(node_id)
        if node is None:
            msg = f"Node {node_id} not found after update"
            raise RuntimeError(msg)
        return node

    async def _delete_node(self, node_id: str) -> bool:
        async with self._connection() as conn:
            sql = (
                "DELETE FROM "
                + _safe_ident(self._settings.edges_table)
                + " WHERE from_node = :id OR to_node = :id"
            )  # nosec B608
            await conn.execute(
                text(sql),
                {"id": node_id},
            )
            sql2 = (
                "DELETE FROM "
                + _safe_ident(self._settings.nodes_table)
                + " WHERE id = :id"
            )  # nosec B608
            result = await conn.execute(
                text(sql2),
                {"id": node_id},
            )
        return result.rowcount > 0

    async def _create_edge(
        self,
        edge_type: str,
        from_node_id: str,
        to_node_id: str,
        properties: dict[str, t.Any],
    ) -> GraphEdgeModel:
        edge_id = str(uuid4())
        now = datetime.now(tz=UTC)
        async with self._connection() as conn:
            sql = (
                "INSERT INTO "
                + _safe_ident(self._settings.edges_table)
                + (
                    " (id, type, from_node, to_node, properties, created_at, updated_at) "
                    "VALUES (:id, :type, :from_node, :to_node, :properties, :created_at, :updated_at)"
                )
            )  # nosec B608
            await conn.execute(
                text(sql),
                {
                    "id": edge_id,
                    "type": edge_type,
                    "from_node": from_node_id,
                    "to_node": to_node_id,
                    "properties": json.dumps(properties),
                    "created_at": now,
                    "updated_at": now,
                },
            )
        return GraphEdgeModel(
            id=edge_id,
            type=edge_type,
            from_node=from_node_id,
            to_node=to_node_id,
            properties=properties,
            created_at=now,
            updated_at=now,
        )

    async def _get_edge(self, edge_id: str) -> GraphEdgeModel | None:
        async with self._connection() as conn:
            sql = (
                "SELECT id, type, from_node, to_node, properties, created_at, updated_at FROM "
                + _safe_ident(self._settings.edges_table)
                + " WHERE id = :id"
            )  # nosec B608
            result = await conn.execute(
                text(sql),
                {"id": edge_id},
            )
            row = result.mappings().first()
        if not row:
            return None
        return GraphEdgeModel(
            id=row["id"],
            type=row["type"],
            from_node=row["from_node"],
            to_node=row["to_node"],
            properties=json.loads(row["properties"]) if row["properties"] else {},
            created_at=row["created_at"],
            updated_at=row["updated_at"],
        )

    async def _update_edge(
        self,
        edge_id: str,
        properties: dict[str, t.Any],
    ) -> GraphEdgeModel:
        now = datetime.now(tz=UTC)
        async with self._connection() as conn:
            sql = (
                "UPDATE "
                + _safe_ident(self._settings.edges_table)
                + " SET properties = :properties, updated_at = :updated_at WHERE id = :id"
            )  # nosec B608
            await conn.execute(
                text(sql),
                {
                    "id": edge_id,
                    "properties": json.dumps(properties),
                    "updated_at": now,
                },
            )
        edge = await self._get_edge(edge_id)
        if edge is None:
            msg = f"Edge {edge_id} not found after update"
            raise RuntimeError(msg)
        return edge

    async def _delete_edge(self, edge_id: str) -> bool:
        async with self._connection() as conn:
            sql = (
                "DELETE FROM "
                + _safe_ident(self._settings.edges_table)
                + " WHERE id = :id"
            )  # nosec B608
            result = await conn.execute(
                text(sql),
                {"id": edge_id},
            )
        return result.rowcount > 0

    async def _find_path(
        self,
        from_node_id: str,
        to_node_id: str,
        max_depth: int | None,
        direction: GraphTraversalDirection,
    ) -> list[GraphPathModel]:
        paths = await self._search_paths(
            from_node_id,
            to_node_id,
            max_depth=max_depth,
            direction=direction,
            shortest_only=False,
        )
        return paths

    async def _find_shortest_path(
        self,
        from_node_id: str,
        to_node_id: str,
        weight_property: str | None,
    ) -> GraphPathModel | None:
        paths = await self._search_paths(
            from_node_id,
            to_node_id,
            max_depth=None,
            direction=GraphTraversalDirection.BOTH,
            weight_property=weight_property,
            shortest_only=True,
        )
        return paths[0] if paths else None

    async def _get_neighbors(
        self,
        node_id: str,
        direction: GraphTraversalDirection,
        edge_types: list[str] | None,
    ) -> list[GraphNodeModel]:
        edges = await self._fetch_edges()
        neighbors: set[str] = set()
        for edge in edges:
            if edge_types and edge["type"] not in edge_types:
                continue
            if direction in (GraphTraversalDirection.OUT, GraphTraversalDirection.BOTH):
                if edge["from_node"] == node_id:
                    neighbors.add(edge["to_node"])
            if direction in (GraphTraversalDirection.IN, GraphTraversalDirection.BOTH):
                if edge["to_node"] == node_id:
                    neighbors.add(edge["from_node"])
        return [
            node
            for node in (await self._get_node_list(list(neighbors)))
            if node is not None
        ]

    async def _get_node_list(self, node_ids: list[str]) -> list[GraphNodeModel | None]:
        if not node_ids:
            return []
        placeholders = ", ".join(f":id_{index}" for index in range(len(node_ids)))
        params = {f"id_{index}": value for index, value in enumerate(node_ids)}
        async with self._connection() as conn:
            sql = (
                "SELECT id, labels, properties, created_at, updated_at FROM "
                + _safe_ident(self._settings.nodes_table)
                + f" WHERE id IN ({placeholders})"
            )  # nosec B608
            result = await conn.execute(
                text(sql),
                params,
            )
            rows = result.mappings().all()
        mapping = {
            row["id"]: GraphNodeModel(
                id=row["id"],
                labels=json.loads(row["labels"]) if row["labels"] else [],
                properties=json.loads(row["properties"]) if row["properties"] else {},
                created_at=row["created_at"],
                updated_at=row["updated_at"],
            )
            for row in rows
        }
        return [mapping.get(node_id) for node_id in node_ids]

    async def _get_schema(self) -> GraphSchemaModel:
        async with self._connection() as conn:
            sql_nodes = (
                "SELECT DISTINCT labels FROM "
                + _safe_ident(self._settings.nodes_table)
                + " WHERE labels IS NOT NULL"
            )  # nosec B608
            node_rows = await conn.execute(text(sql_nodes))
            sql_edges = "SELECT DISTINCT type FROM " + _safe_ident(
                self._settings.edges_table
            )  # nosec B608
            edge_rows = await conn.execute(text(sql_edges))
        node_labels: set[str] = set()
        for row in node_rows.fetchall():
            if row[0]:
                node_labels.update(json.loads(row[0]))
        edge_types = [row[0] for row in edge_rows.fetchall() if row[0]]
        return GraphSchemaModel(
            node_types=sorted(node_labels),
            edge_types=sorted(set(edge_types)),
            constraints=[],
            indexes=[],
        )

    async def _create_index(
        self,
        labels: list[str],
        properties: list[str],
        index_type: str,
    ) -> bool:
        self.logger.warning(
            "Custom index creation is not supported for DuckDB PGQ adapter"
        )
        return False

    async def _drop_index(self, index_name: str) -> bool:
        self.logger.warning(
            "Custom index removal is not supported for DuckDB PGQ adapter"
        )
        return False

    async def _bulk_create_nodes(
        self,
        nodes: list[dict[str, t.Any]],
    ) -> list[GraphNodeModel]:
        return [
            await self._create_node(
                node.get("labels", []),
                node.get("properties", {}),
            )
            for node in nodes
        ]

    async def _bulk_create_edges(
        self,
        edges: list[dict[str, t.Any]],
    ) -> list[GraphEdgeModel]:
        return [
            await self._create_edge(
                edge.get("type", "RELATES"),
                edge["from_node"],
                edge["to_node"],
                edge.get("properties", {}),
            )
            for edge in edges
        ]

    async def _count_nodes(self, labels: list[str] | None) -> int:
        async with self._connection() as conn:
            result = await conn.execute(
                text(f"SELECT labels FROM {self._settings.nodes_table}")
            )
            rows = result.fetchall()
        if not labels:
            return len(rows)
        label_set = set(labels)
        count = 0
        for row in rows:
            raw = row[0]
            if not raw:
                continue
            node_labels = set(json.loads(raw))
            if label_set.issubset(node_labels):
                count += 1
        return count

    async def _count_edges(self, edge_types: list[str] | None) -> int:
        edges = await self._fetch_edges()
        if not edge_types:
            return len(edges)
        edge_type_set = set(edge_types)
        return sum(1 for edge in edges if edge["type"] in edge_type_set)

    async def _clear_graph(self) -> bool:
        async with self._connection() as conn:
            sql_edges = "DELETE FROM " + _safe_ident(self._settings.edges_table)  # nosec B608
            sql_nodes = "DELETE FROM " + _safe_ident(self._settings.nodes_table)  # nosec B608
            await conn.execute(text(sql_edges))
            await conn.execute(text(sql_nodes))
        return True

    async def _search_paths(
        self,
        start: str,
        goal: str,
        *,
        max_depth: int | None,
        direction: GraphTraversalDirection,
        weight_property: str | None = None,
        shortest_only: bool = False,
    ) -> list[GraphPathModel]:
        if start == goal:
            node = await self._get_node(start)
            if node is None:
                return []
            return [
                GraphPathModel(
                    nodes=[node],
                    edges=[],
                    length=0,
                    weight=0.0,
                )
            ]

        edges = await self._fetch_edges()
        adjacency = self._build_adjacency_list(edges, direction)
        found_paths = self._bfs_paths(adjacency, start, goal, max_depth, shortest_only)
        return await self._paths_to_models(found_paths, weight_property)

    def _build_adjacency_list(
        self, edges: list[dict[str, t.Any]], direction: GraphTraversalDirection
    ) -> dict[str, list[dict[str, t.Any]]]:
        adjacency: dict[str, list[dict[str, t.Any]]] = {}
        for edge in edges:
            adjacency.setdefault(edge["from_node"], []).append(edge)
            if direction in (GraphTraversalDirection.IN, GraphTraversalDirection.BOTH):
                reverse = edge.copy()
                reverse["from_node"] = edge["to_node"]
                reverse["to_node"] = edge["from_node"]
                adjacency.setdefault(reverse["from_node"], []).append(reverse)
        return adjacency

    def _bfs_paths(
        self,
        adjacency: dict[str, list[dict[str, t.Any]]],
        start: str,
        goal: str,
        max_depth: int | None,
        shortest_only: bool,
    ) -> list[list[dict[str, str | None]]]:
        queue: list[list[dict[str, str | None]]] = [[{"node": start, "edge": None}]]
        visited: set[str] = {start}
        found_paths: list[list[dict[str, str | None]]] = []
        while queue:
            path = queue.pop(0)
            current: str = path[-1]["node"]  # type: ignore[assignment]
            if current == goal:
                found_paths.append(path)
                if shortest_only:
                    break
                continue
            if max_depth is not None and len(path) - 1 >= max_depth:
                continue
            for edge in adjacency.get(current, []):
                neighbor = edge["to_node"]
                if neighbor not in visited or neighbor == goal:
                    new_path = path + [{"node": neighbor, "edge": edge["id"]}]
                    queue.append(new_path)
                    visited.add(neighbor)
        return found_paths

    async def _paths_to_models(
        self,
        found_paths: list[list[dict[str, str | None]]],
        weight_property: str | None,
    ) -> list[GraphPathModel]:
        graph_paths: list[GraphPathModel] = []
        for raw_path in found_paths:
            node_ids = [step["node"] for step in raw_path if step["node"] is not None]
            nodes = await self._get_node_list(node_ids)
            edge_ids = [step["edge"] for step in raw_path if step["edge"]]
            edge_models = [
                await self._get_edge(edge_id) for edge_id in edge_ids if edge_id
            ]
            weight = None
            if weight_property and edge_models:
                weight = sum(
                    e.properties.get(weight_property, 1)
                    for e in edge_models
                    if e is not None
                )
            graph_paths.append(
                GraphPathModel(
                    nodes=[n for n in nodes if n is not None],
                    edges=[e for e in edge_models if e is not None],
                    length=len(edge_models),
                    weight=weight,
                )
            )
        return graph_paths

    async def _fetch_edges(self) -> list[dict[str, t.Any]]:
        async with self._connection() as conn:
            sql = "SELECT id, type, from_node, to_node, properties FROM " + _safe_ident(
                self._settings.edges_table
            )  # nosec B608
            result = await conn.execute(text(sql))
            rows = result.mappings().all()
        return [
            {
                "id": row["id"],
                "type": row["type"],
                "from_node": row["from_node"],
                "to_node": row["to_node"],
                "properties": json.loads(row["properties"])
                if row["properties"]
                else {},
            }
            for row in rows
        ]


depends.set(Graph, "duckpgq")



"""ACB MCP Serialization Utilities.

Provides functions for serializing and formatting component information for MCP transmission.
"""

from typing import Any


def serialize_component_info(component: Any) -> dict[str, Any]:
    """Serialize component information for MCP transmission."""
    try:
        # Get basic component information
        info: dict[str, Any] = {
            "type": type(component).__name__,
            "module": type(component).__module__,
            "methods": [],
            "attributes": [],
        }

        # Get public methods
        for attr_name in dir(component):
            if not attr_name.startswith("_"):
                attr = getattr(component, attr_name)
                if callable(attr):
                    info["methods"].append(attr_name)
                else:
                    info["attributes"].append(attr_name)

        return info
    except Exception as e:
        return {"type": "Unknown", "error": str(e)}


def format_tool_response(response: Any) -> dict[str, Any]:
    """Format a tool response for MCP transmission."""
    if isinstance(response, dict):
        return response
    if isinstance(response, list | tuple):
        return {"items": list(response)}
    if isinstance(response, str | int | float | bool):
        return {"value": response}
    return {"value": str(response)}
